from typing import Union, Dict, List, Tuple, Callable, Any
from torch.distributions import Categorical
import torch as t
import torch.nn as nn

from machin.frame.buffers.buffer import Transition, Buffer
from machin.frame.noise.action_space_noise import \
    add_normal_noise_to_action, \
    add_clipped_normal_noise_to_action, \
    add_uniform_noise_to_action, \
    add_ou_noise_to_action
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework
from .utils import hard_update, soft_update, safe_call, assert_output_is_probs


class DDPG(TorchFramework):
    """
    DDPG framework.
    """

    _is_top = ["actor", "critic", "actor_target", "critic_target"]
    _is_restorable = ["actor_target", "critic_target"]

    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 actor_target: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 action_trans_func: Callable = None,
                 visualize: bool = False,
                 **__):
        """
        Note:
            Your optimizer will be called as::

                optimizer(network.parameters(), learning_rate)

            Your lr_scheduler will be called as::

                lr_scheduler(
                    optimizer,
                    *lr_scheduler_args[0],
                    **lr_scheduler_kwargs[0],
                )

            Your criterion will be called as::

                criterion(
                    target_value.view(batch_size, 1),
                    predicted_value.view(batch_size, 1)
                )

        Note:
            The action transform function is used to transform the output
            of actor to the input of critic.
            Action transform function must accept:

              1. Raw action from the actor model.
              2. Concatenated :attr:`.Transition.next_state`.
              3. Any other concatenated lists of custom keys from \
                  :class:`.Transition`.

            and returns:

              1. Something with the same form as :attr:`.Transition.action`

            by default it is :meth:`.DDPG.action_transform_function`

        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            reward_func: Reward function used in training.
            action_trans_func: Action transform function, used to transform
                the raw output of your actor.
            visualize: Whether visualize the network flow in the first pass.
        """
        self.batch_size = batch_size
        self.update_rate = update_rate
        self.discount = discount
        self.visualize = visualize

        self.actor = actor
        self.actor_target = actor_target
        self.critic = critic
        self.critic_target = critic_target
        self.actor_optim = optimizer(self.actor.parameters(),
                                     lr=learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(),
                                      lr=learning_rate)
        self.replay_buffer = (Buffer(replay_size, replay_device)
                              if replay_buffer is None
                              else replay_buffer)

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.actor, self.actor_target)
            hard_update(self.critic, self.critic_target)

        if lr_scheduler is not None:
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim,
                *lr_scheduler_args[0],
                **lr_scheduler_kwargs[0]
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim,
                *lr_scheduler_args[1],
                **lr_scheduler_kwargs[1]
            )

        self.criterion = criterion

        self.reward_func = (DDPG.bellman_function
                            if reward_func is None
                            else reward_func)
        self.action_transform_func = (DDPG.action_transform_function
                                      if action_trans_func is None
                                      else action_trans_func)

        super(DDPG, self).__init__()

    def act(self,
            state: Dict[str, Any],
            use_target: bool = False,
            **__):
        """
        Use actor network to produce an action for the current state.

        Args:
            state: Current state.
            use_target: Whether use the target network.

        Returns:
            Action of shape ``[batch_size, action_dim]``.
        """
        if use_target:
            return safe_call(self.actor_target, state)
        else:
            return safe_call(self.actor, state)

    def act_with_noise(self,
                       state: Dict[str, Any],
                       noise_param: Tuple = (0.0, 1.0),
                       ratio: float = 1.0,
                       mode: str = "uniform",
                       use_target: bool = False,
                       **__):
        """
        Use actor network to produce a noisy action for the current state.

        See Also:
             :mod:`machin.frame.noise.action_space_noise`

        Args:
            state: Current state.
            noise_param: Noise params.
            ratio: Noise ratio.
            mode: Noise mode. Supported are:
                ``"uniform", "normal", "clipped_normal", "ou"``
            use_target: Whether use the target network.

        Returns:
            Noisy action of shape ``[batch_size, action_dim]``.
        """
        if mode == "uniform":
            return add_uniform_noise_to_action(
                self.act(state, use_target), noise_param, ratio
            )
        if mode == "normal":
            return add_normal_noise_to_action(
                self.act(state, use_target), noise_param, ratio
            )
        if mode == "clipped_normal":
            return add_clipped_normal_noise_to_action(
                self.act(state, use_target), noise_param, ratio
            )
        if mode == "ou":
            return add_ou_noise_to_action(
                self.act(state, use_target), noise_param, ratio
            )
        raise RuntimeError("Unknown noise type: " + str(mode))

    def act_discreet(self,
                     state: Dict[str, Any],
                     use_target: bool = False,
                     **__):
        """
        Use actor network to produce a discreet action for the current state.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Action of shape ``[batch_size, 1]``.
        """
        if use_target:
            result = safe_call(self.actor_target, state)
        else:
            result = safe_call(self.actor, state)

        assert_output_is_probs(result)
        batch_size = result.shape[0]
        result = t.argmax(result, dim=1).view(batch_size, 1)
        return result

    def act_discreet_with_noise(self,
                                state: Dict[str, Any],
                                use_target: bool = False,
                                **__):
        """
        Use actor network to produce a noisy discreet action for
        the current state.

        Notes:
            actor network must output a probability tensor, of shape
            (batch_size, action_dims), and has a sum of 1 for each row
            in dimension 1.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Noisy action of shape ``[batch_size, 1]``.
        """
        if use_target:
            result = safe_call(self.actor_target, state)
        else:
            result = safe_call(self.actor, state)

        assert_output_is_probs(result)
        dist = Categorical(result)
        batch_size = result.shape[0]
        return dist.sample([batch_size, 1])

    def criticize(self,
                  state: Dict[str, Any],
                  action: Dict[str, Any],
                  use_target: bool = False,
                  **__):
        """
        Use critic network to evaluate current value.

        Args:
            state: Current state.
            action: Current action.
            use_target: Whether to use the target network.

        Returns:
            Value of shape ``[batch_size, 1]``.
        """
        if use_target:
            return safe_call(self.critic_target, state, action)
        else:
            return safe_call(self.critic, state, action)

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.
        """
        self.replay_buffer.append(transition, required_attrs=(
            "state", "action", "reward", "next_state", "terminal"
        ))

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        for trans in episode:
            self.replay_buffer.append(trans, required_attrs=(
                "state", "action", "reward", "next_state", "terminal"
            ))

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.
            update_target: Whether to update targets.
            concatenate_samples: Whether to concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """
        batch_size, (state, action, reward, next_state, terminal, *others) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        # Update critic network first.
        # Generate value reference :math: `y_i` using target actor and
        # target critic.
        with t.no_grad():
            next_action = self.action_transform_func(self.act(next_state, True),
                                                     next_state,
                                                     *others)
            next_value = self.criticize(next_state, next_action, True)
            next_value = next_value.view(batch_size, -1)
            y_i = self.reward_func(reward, self.discount, next_value,
                                   terminal, *others)

        cur_value = self.criticize(state, action)
        value_loss = self.criterion(cur_value, y_i.to(cur_value.device))

        if self.visualize:
            self.visualize_model(value_loss, "critic")

        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            self.critic_optim.step()

        # Update actor network
        cur_action = self.action_transform_func(self.act(state), state, *others)
        act_value = self.criticize(state, cur_action)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor")

        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            self.actor_optim.step()

        # Update target networks
        if update_target:
            soft_update(self.actor_target, self.actor, self.update_rate)
            soft_update(self.critic_target, self.critic, self.update_rate)

        # use .item() to prevent memory leakage
        return -act_policy_loss.item(), value_loss.item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()

    def load(self, model_dir: str, network_map: Dict[str, str] = None,
             version: int = -1):
        # DOC INHERITED
        super(DDPG, self).load(model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.actor, self.actor_target)
            hard_update(self.critic, self.critic_target)

    @staticmethod
    def action_transform_function(raw_output_action: Any, *_):
        return {"action": raw_output_action}

    @staticmethod
    def bellman_function(reward, discount, next_value, terminal, *_):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * (1 - terminal) * next_value



from .a2c import *
from machin.parallel.server import PushPullGradServer
from machin.parallel.distributed import RpcGroup
from machin.utils.helper_classes import Switch


class A3C(A2C):
    """
    A3C framework.
    """
    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 grad_server_group: RpcGroup = None,
                 grad_servers: Tuple[Any, Any] = None,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 learning_rate: float = 0.001,
                 entropy_weight: float = None,
                 value_weight: float = 0.5,
                 gradient_max: float = np.inf,
                 gae_lambda: float = 1.0,
                 discount: float = 0.99,
                 update_times: int = 50,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 visualize: bool = False,
                 **__):
        """
        See Also:
            :class:`.A2C`

        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            grad_server_group: Gradient sync server group, actor and critic will
                share the same sync group.
            grad_servers: Custom gradient sync servers, the first server is for
                actor, and the second one is for critic.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            gradient_max: Maximum gradient.
            gae_lambda: :math:`\\lambda` used in generalized advantage
                estimation.
            discount: :math:`\\gamma` used in the bellman function.
            update_times: Number of update iterations per sample period. Buffer
                will be cleared after ``update()``
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
        """
        super(A3C, self).__init__(actor, critic, optimizer, criterion,
                                  lr_scheduler=lr_scheduler,
                                  lr_scheduler_args=lr_scheduler_args,
                                  lr_scheduler_kwargs=lr_scheduler_kwargs,
                                  learning_rate=learning_rate,
                                  entropy_weight=entropy_weight,
                                  value_weight=value_weight,
                                  gradient_max=gradient_max,
                                  gae_lambda=gae_lambda,
                                  discount=discount,
                                  update_times=update_times,
                                  replay_size=replay_size,
                                  replay_device=replay_device,
                                  replay_buffer=replay_buffer,
                                  visualize=visualize)
        if grad_servers is None:
            self.actor_grad_server = PushPullGradServer(
                group=grad_server_group
            )
            self.critic_grad_server = PushPullGradServer(
                group=grad_server_group
            )
        else:
            self.actor_grad_server, self.critic_grad_server = \
                grad_servers[0], grad_servers[1]
        self._disable_sync = Switch()

    def act(self, state: Dict[str, Any], pull: bool = True, **__):
        # DOC INHERITED
        if pull and not self._disable_sync.get():
            self.actor_grad_server.pull(self.actor)
        return safe_call(self.actor, state)

    def eval_act(self,
                 state: Dict[str, Any],
                 action: Dict[str, Any],
                 pull: bool = True,
                 **__):
        # DOC INHERITED
        if pull and not self._disable_sync.get():
            self.actor_grad_server.pull(self.actor)
        return safe_call(self.actor, state, action)

    def criticize(self, state: Dict[str, Any], *_, pull=True, **__):
        # DOC INHERITED
        if pull and not self._disable_sync.get():
            self.critic_grad_server.pull(self.critic)
        return safe_call(self.critic, state)

    def update(self,
               update_value=True,
               update_policy=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        self._disable_sync.on()
        self.update(update_value, update_policy, concatenate_samples)
        self._disable_sync.off()
        self.actor_grad_server.push(self.actor)
        self.critic_grad_server.push(self.critic)



from abc import ABC, abstractmethod
from typing import Any
import torch as t
import torch.distributions as tdist

from machin.utils.checker import check_shape


class NoiseGen(ABC):
    """
    Base class for noise generators.
    """
    @abstractmethod
    def __call__(self, device=None):
        """
        Generate a noise tensor, and move it to the specified device.
        """
        pass

    @abstractmethod
    def __repr__(self):
        """
        Return a correct representation of the noise distribution, must
        conform to style: "<NoiseName>(param1=..., param2=...)".
        """
        return "EmptyNoise()"

    def reset(self):
        """
        Reset internal states of the noise generator, if it has any.
        """
        pass


class NormalNoiseGen(NoiseGen):
    def __init__(self, shape: Any, mu: float = 0.0, sigma: float = 1.0):
        """
        Normal noise generator.

        Example:
            >>> gen = NormalNoiseGen([2, 3], 0, 1)
            >>> gen("cuda:0")
            tensor([[-0.5957,  0.2360,  1.0999],
                    [ 1.6259,  1.2052, -0.0667]], device='cuda:0')

        Args:
            shape: Output shape.
            mu: Average mean of normal noise.
            sigma: Standard deviation of normal noise.
        """
        self.mu = mu
        self.sigma = sigma
        self.dist = tdist.normal.Normal(mu, sigma)
        self.shape = shape

    def __call__(self, device=None):
        if device is not None:
            return self.dist.sample(self.shape).to(device)
        else:
            return self.dist.sample(self.shape)

    def __repr__(self):
        return 'NormalNoise(mu={}, sigma={})'.format(self.mu, self.sigma)


class UniformNoiseGen(NoiseGen):
    def __init__(self, shape: Any, umin: float = 0.0, umax: float = 1.0):
        """
        Normal noise generator.

        Example:
            >>> gen = UniformNoiseGen([2, 3], 0, 1)
            >>> gen("cuda:0")
            tensor([[0.0745, 0.6581, 0.9572],
                    [0.4450, 0.8157, 0.6421]], device='cuda:0')

        Args:
            shape: Output shape.
            umin: Minimum value of uniform noise.
            umax: Maximum value of uniform noise.
        """
        self.min = umin
        self.max = umax
        self.dist = tdist.uniform.Uniform(umin, umax)
        self.shape = shape

    def __call__(self, device=None):
        if device is not None:
            return self.dist.sample(self.shape).to(device)
        else:
            return self.dist.sample(self.shape)

    def __repr__(self):
        return 'UniformNoise(min={}, max={})'.format(self.min, self.max)


class OrnsteinUhlenbeckNoiseGen(NoiseGen):
    def __init__(self, shape: Any, mu: float = 0.0, sigma: float = 1.0,
                 theta: float = 0.15, dt: float = 1e-2,
                 x0: t.Tensor = None):
        """
        Ornstein-Uhlenbeck noise generator.
        Based on `definition <http://math.stackexchange.com/questions\
/1287634/implementing-ornstein-uhlenbeck-in-matlab>`_:

        :math:`X_{n+1} = X_n + \\theta (\\mu - X_n)\\Delta t + \\sigma \
        \\Delta W_n`

        Example:
            >>> gen = OrnsteinUhlenbeckNoiseGen([2, 3], 0, 1)
            >>> gen("cuda:0")
            tensor([[ 0.1829,  0.1589, -0.1932],
                    [-0.1568,  0.0579,  0.2107]], device='cuda:0')
            >>> gen.reset()


        Args:
            shape: Output shape.
            mu: Average mean of noise.
            sigma: Weight of the random wiener process.
            theta: Weight of difference correction.
            dt: Time step size.
            x0: Initial x value. Must have the same shape as ``shape``.
        """
        self.theta = theta
        self.mu = mu
        self.sigma = sigma
        self.dt = t.tensor(dt) if not isinstance(dt, t.Tensor) else dt
        self.norm_dist = tdist.normal.Normal(loc=0.0, scale=1.0)
        self.shape = shape
        self.x0 = x0
        check_shape(x0, list(shape))
        self.x_prev = None
        self.reset()

    def __call__(self, device=None):
        x = self.x_prev + self.theta * (self.mu - self.x_prev) * self.dt + \
            self.sigma * t.sqrt(self.dt) * self.norm_dist.sample(self.shape)
        self.x_prev = x
        if device is not None:
            return x.to(device)
        else:
            return x

    def reset(self):
        """
        Reset the generator to its initial state.
        """
        self.x_prev = self.x0 if self.x0 is not None else t.zeros(self.shape)

    def __repr__(self):
        return 'OrnsteinUhlenbeckNoise(mu={}, sigma={})'\
            .format(self.mu, self.sigma)



from datetime import datetime, timedelta
from typing import Union, Iterable

import os

from machin.utils.prepare import \
    prep_dirs_default, prep_clear_dirs, prep_create_dirs


class SaveEnv:
    def __init__(self,
                 env_root: str,
                 env_dirs: Union[dict, None] = None,
                 restart_use_trial: Union[str, None] = None,
                 time_format="%Y_%m_%d_%H_%M_%S"):
        """
        Create the default environment for saving. creates something like::

            <your environment root>
            ├── config
            ├── log
            │   ├── images
            │   └── train_log
            └── model

        Args:
            env_root: root directory for all trials of the environment.
            env_dirs: directory mapping for sub directories, such as
                log, model, etc.
            restart_use_trial: in format ``time_format``
        """
        self.env_root = env_root
        self.env_dirs = env_dirs
        self.time_format = time_format

        if restart_use_trial is None:
            self.env_create_time = datetime.now()
        else:
            self.env_create_time = datetime.strptime(restart_use_trial,
                                                     self.time_format)
        self._check_dirs()
        self._prep_dirs()

    def create_dirs(self, dirs: Iterable[str]):
        """
        Create additional directories in root.

        Args:
            dirs: Directories.
        """
        prep_create_dirs([os.path.join(self.env_root, dir) for dir in dirs])

    def get_trial_root(self):
        # pylint: disable=missing-docstring
        return os.path.join(self.env_root,
                            self.env_create_time.strftime(self.time_format))

    def get_trial_config_file(self):
        # pylint: disable=missing-docstring
        return os.path.join(self.env_root,
                            self.env_create_time.strftime(self.time_format),
                            "config/config.json")

    def get_trial_model_dir(self):
        # pylint: disable=missing-docstring
        return os.path.join(self.env_root,
                            self.env_create_time.strftime(self.time_format),
                            "model")

    def get_trial_image_dir(self):
        # pylint: disable=missing-docstring
        return os.path.join(self.env_root,
                            self.env_create_time.strftime(self.time_format),
                            "log/images")

    def get_trial_train_log_dir(self):
        # pylint: disable=missing-docstring
        return os.path.join(self.env_root,
                            self.env_create_time.strftime(self.time_format),
                            "log/train_log")

    def get_trial_time(self):
        # pylint: disable=missing-docstring
        return self.env_create_time

    def clear_trial_config_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs([
            os.path.join(self.env_root,
                         self.env_create_time.strftime(self.time_format),
                         "config")
        ])

    def clear_trial_model_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs([
            os.path.join(self.env_root,
                         self.env_create_time.strftime(self.time_format),
                         "model")
        ])

    def clear_trial_image_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs([
            os.path.join(self.env_root,
                         self.env_create_time.strftime(self.time_format),
                         "log/images")
        ])

    def clear_trial_train_log_dir(self):
        # pylint: disable=missing-docstring
        prep_clear_dirs([
            os.path.join(self.env_root,
                         self.env_create_time.strftime(self.time_format),
                         "log/train_log")
        ])

    def remove_trials_older_than(self,
                                 diff_day: int = 0,
                                 diff_hour: int = 1,
                                 diff_minute: int = 0,
                                 diff_second: int = 0):
        """
        By default this function removes all trials started one hour earlier
        than current time.

        Args:
            diff_day: Difference in days.
            diff_hour: Difference in hours.
            diff_minute: Difference in minutes.
            diff_second: Difference in seconds.
        """
        trial_list = [f for f in os.listdir(self.env_root)]
        current_time = datetime.now()
        diff_threshold = timedelta(days=diff_day, hours=diff_hour,
                                   minutes=diff_minute, seconds=diff_second)
        for file in trial_list:
            try:
                time = datetime.strptime(file, self.time_format)
            except ValueError:
                # not a trial
                pass
            else:
                diff_time = current_time - time
                if diff_time > diff_threshold:
                    rm_path = os.path.join(self.env_root, file)
                    print("Removing trial directory: {}".format(rm_path))
                    os.remove(rm_path)

    def _prep_dirs(self):
        root_dir = os.path.join(self.env_root,
                                self.env_create_time.strftime(self.time_format))
        prep_dirs_default(root_dir)

    def _check_dirs(self):
        """
        Overload this function in your environment class to check directory
        mapping

        Raises:
            RuntimeError if directory mapping is invalid.
        """
        pass



from typing import Dict, Iterable, Any

import os
import re
import torch as t
import torch.nn as nn

from .conf import Config, get_args, merge_config, \
    replace_config, save_config, load_config_dict


def prep_args(config: Config, save_env):
    """
    Args:
        config: :class:`.Config` object to prepare.
        save_env: :class:`.SaveEnv` instance.

    Example::

        from machin.utils.conf import Config
        from machin.utils.save_env import SaveEnv
        c = Config()

        # set some config attributes
        # restart_from_trial specifies the trial name in your root
        # directory, if it is set, then prep_args will load arguments
        # from that trial record.

        c.model_save_int = 100
        c.some_root_dir = "/some_directory"
        c.restart_from_trial = "2020_05_09_15_00_31"

        save_env = SaveEnv(c.some_root_dir,
                           restart_use_trial=c.restart_from_trial)

        # If c.restart_from_trial is not None, config will be
        # overwritten by the config entries in the saved config
        # of trial.
        # Otherwise, it is saved in the directory:
        # <c.some_root_dir>/<trial_start_time>/config
        # as a json file.
        prep_args(c, save_env)
    """
    c = config
    args = get_args()
    merge_config(c, args.conf)

    # preparations
    if c.restart_from_trial is not None:
        r = c.restart_from_trial
        replace_config(c, load_config_dict(save_env.get_trial_config_file()))
        save_env.clear_trial_train_log_dir()
        # prevent overwriting
        c.restart_from_trial = r
    else:
        save_config(c, save_env.get_trial_config_file())


def prep_clear_dirs(dirs: Iterable[str]):
    """
    Args:
         dirs: a list of directories to clear
    """
    for dir_ in dirs:
        file_list = [f for f in os.listdir(dir_)]
        for f in file_list:
            os.remove(os.path.join(dir_, f))


def prep_create_dirs(dirs: Iterable[str]):
    """
    Note: will recursively create directories.

    Args:
        dirs: a list of directories to create if these directories
            are not found.
    """
    for dir_ in dirs:
        if not os.path.exists(dir_):
            os.makedirs(dir_)


def prep_dirs_default(root_dir, clear_old=False):
    """
    Create default directories (/model, /log/images, /log/train_log) in the
    given root directory.

    Args:
        root_dir: root directory.
        clear_old: whether completely removes all things in the old
            default directories.
    """
    prep_create_dirs((root_dir + "/model",
                      root_dir + "/config",
                      root_dir + "/log/images",
                      root_dir + "/log/train_log"))
    if clear_old:
        prep_clear_dirs((root_dir + "/model",
                         root_dir + "/config",
                         root_dir + "/log/images",
                         root_dir + "/log/train_log"))


def prep_load_state_dict(model: nn.Module,
                         state_dict: Any):
    """
    Automatically load a **loaded state dictionary**

    Note:
        This function handles tensor device remapping.
    """
    for name, param in model.named_parameters():
        if name not in state_dict:
            print("Warning: Key {} not found in state dict.".format(name))
        state_dict[name].to(param.device)
    model.load_state_dict(state_dict)


def prep_load_model(model_dir: str,
                    model_map: Dict[str, str],
                    version: int = -1,
                    quiet: bool = False):
    """
    Automatically find and load models.

    Args:
        model_dir: Directory to save models.
        model_map: Model saving map.
        version: Version to load, if specified, otherwise automatically
            find the latest version.
        quiet: Raise no error if no valid version could be found.
    """
    if not os.path.exists(model_dir):
        raise RuntimeError("Model directory doesn't exist!")
    version_map = {}
    for net_name in model_map.keys():
        version_map[net_name] = set()
    models = os.listdir(model_dir)
    for m in models:
        match = re.fullmatch("([a-zA-Z0-9_-]+)_([0-9]+)\\.pt$", m)
        if match is not None:
            n = match.group(1)
            v = int(match.group(2))
            if n in model_map:
                version_map[n].add(v)
    if version > 0:
        is_version_found = [version in version_map[name]
                            for name in model_map.keys()]
        if all(is_version_found):
            print("Specified version found, using version: {}".format(version))
            for net_name, net in model_map.items():
                net = net  # type: nn.Module
                state_dict = t.load(model_dir +
                                    "/{}_{}.pt".format(net_name, version),
                                    map_location="cpu")
                prep_load_state_dict(net, state_dict)
            return
        else:
            for ivf, net_name in zip(is_version_found, model_map.keys()):
                if not ivf:
                    print("Specified version {} for network {} is invalid"
                          .format(version, net_name))

    print("Begin auto find")
    # use the valid, latest model
    common = set.intersection(*version_map.values())
    if len(common) == 0:
        if not quiet:
            raise RuntimeError("Cannot find a valid version for all models!")
        else:
            return
    version = max(common)
    print("Using version: {}".format(version))
    for net_name, net in model_map.items():
        state_dict = t.load(model_dir +
                            "/{}_{}.pt".format(net_name, version),
                            map_location="cpu")
        prep_load_state_dict(net, state_dict)



from .election import \
    ElectionGroupBase, \
    ElectionGroupSimple, \
    ElectionGroupStableBase, \
    ElectionGroupStableRpc
from .role import RoleBase
from .role_dispatcher import \
    RoleDispatcherBase, \
    RoleDispatcherSimple, \
    RoleDispatcherElection
from .world import World, CollectiveGroup, RpcGroup

from . import election
from . import role
from . import role_dispatcher
from . import world

__all__ = [
    "ElectionGroupBase",
    "ElectionGroupSimple",
    "ElectionGroupStableBase",
    "ElectionGroupStableRpc",
    "RoleBase",
    "RoleDispatcherBase",
    "RoleDispatcherSimple",
    "RoleDispatcherElection",
    "World", "CollectiveGroup", "RpcGroup",
    "election", "role", "role_dispatcher", "world"
]



import multiprocessing.pool as pool
from torch.multiprocessing.pool import clean_worker

import dill
from .queue import SimpleQueue


def proxy_caller(input_):
    """
    Call a serialized function and return results.
    """
    func_str, args, kwargs = input_
    func = dill.loads(func_str)
    return func(*args, **kwargs)


def proxy_dumper(recurse, func, args_list, kwargs_list=None):
    """
    Serialize a function so it can be called.

    Returns:
        List[function string, arguments...]
    """
    # recurse will enable context variable saving
    dump = dill.dumps(func, protocol=dill.HIGHEST_PROTOCOL, recurse=recurse)
    if kwargs_list is not None:
        for args, kwargs in zip(args_list, kwargs_list):
            yield [dump, args, kwargs]
    else:
        for args in args_list:
            yield [dump, args, {}]


class Pool(pool.Pool):
    """
    Enhanced multiprocessing pool for pytorch, provides:
     1. Support for lambdas and local functions.
     2. Ability to select the tensor serialize scheme.
     3. Ability to get the number of workers.
    """

    # Multiprocessing pool is badly written.
    # python IDEs will complain a lot.

    def __init__(self, processes=None, initializer=None, initargs=(),
                 maxtasksperchild=None, context=None,
                 is_global=True, is_daemon=True, is_copy_tensors=True):
        self.is_global = is_global
        self.is_daemon = is_daemon
        self.is_copy_tensors = is_copy_tensors
        super(Pool, self).__init__(
            processes=processes,
            initializer=initializer,
            initargs=initargs,
            maxtasksperchild=maxtasksperchild,
            context=context
        )

    def _setup_queues(self):
        self._inqueue = SimpleQueue(ctx=self._ctx,
                                    copy_tensor=self.is_copy_tensors)
        self._outqueue = SimpleQueue(ctx=self._ctx,
                                     copy_tensor=self.is_copy_tensors)
        self._quick_put = self._inqueue.quick_put
        self._quick_get = self._outqueue.quick_get

    def apply(self, func, args=(), kwds=None):
        # DOC INHERITED
        if kwds is None:
            kwds = {}
        return pool.Pool.apply(self, proxy_caller,
                               [dill.dumps(func, recurse=self.is_global),
                                args, kwds])

    def apply_async(self, func, args=(), kwds=None, callback=None,
                    error_callback=None):
        # DOC INHERITED
        if kwds is None:
            kwds = {}
        return pool.Pool.apply_async(self, proxy_caller,
                                     [dill.dumps(func, recurse=self.is_global),
                                      args, kwds])

    def map(self, func, iterable, chunksize=None):
        # DOC INHERITED
        return pool.Pool.map(self, proxy_caller,
                             proxy_dumper(
                                 self.is_global,
                                 func,
                                 [(arg,) for arg in iterable]
                             ),
                             chunksize)

    def map_async(self, func, iterable, chunksize=None, callback=None,
                  error_callback=None):
        # DOC INHERITED
        return pool.Pool.map_async(self, proxy_caller,
                                   proxy_dumper(
                                       self.is_global,
                                       func,
                                       [(arg,) for arg in iterable]
                                   ),
                                   chunksize, callback, error_callback)

    def imap(self, func, iterable, chunksize=1):
        # DOC INHERITED
        return pool.Pool.imap(self, proxy_caller,
                              proxy_dumper(
                                  self.is_global,
                                  func,
                                  [(arg,) for arg in iterable]
                              ),
                              chunksize)

    def imap_unordered(self, func, iterable, chunksize=1):
        # DOC INHERITED
        return pool.Pool.imap_unordered(self, proxy_caller,
                                        proxy_dumper(
                                            self.is_global,
                                            func,
                                            [(arg,) for arg in iterable]
                                        ),
                                        chunksize)

    def starmap(self, func, iterable, chunksize=None):
        # DOC INHERITED
        return pool.Pool.starmap(self, proxy_caller,
                                 proxy_dumper(
                                     self.is_global,
                                     func,
                                     iterable
                                 ),
                                 chunksize)

    def starmap_async(self, func, iterable, chunksize=None, callback=None,
                      error_callback=None):
        # DOC INHERITED
        return pool.Pool.starmap_async(self, proxy_caller,
                                       proxy_dumper(
                                           self.is_global,
                                           func,
                                           iterable
                                       ),
                                       chunksize, callback, error_callback)

    def _repopulate_pool(self):
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for _ in range(self._processes - len(self._pool)):
            # changed worker -> clean_worker
            args = (self._inqueue, self._outqueue,
                    self._initializer,
                    self._initargs, self._maxtasksperchild)
            if hasattr(self, '_wrap_exception'):
                args += (self._wrap_exception,)
            worker = self.Process(target=clean_worker, args=args)
            self._pool.append(worker)
            worker.name = worker.name.replace('Process', 'PoolWorker')
            worker.daemon = self.is_daemon
            worker.start()
            pool.util.debug('added worker')

    def size(self):
        """
        Returns:
            The number of workers in pool.
        """
        return len(self._pool)

    def __reduce__(self):
        raise RuntimeError("Process pool is not reducible.")


class ThreadPool(pool.ThreadPool):
    """
    A typical thread pool, provides:
    1. Ability to get the number of workers.
    """
    # Multiprocessing pool is badly written.
    # python IDEs will complain a lot.

    def _repopulate_pool(self):
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for _ in range(self._processes - len(self._pool)):
            # changed worker -> clean_worker
            args = (self._inqueue, self._outqueue,
                    self._initializer,
                    self._initargs, self._maxtasksperchild)
            if hasattr(self, '_wrap_exception'):
                args += (self._wrap_exception,)
            worker = self.Process(target=clean_worker, args=args)
            self._pool.append(worker)
            worker.name = worker.name.replace('Process', 'PoolWorker')
            worker.start()
            pool.util.debug('added worker')

    def size(self):
        """
        Returns:
            The number of workers in pool.
        """
        return len(self._pool)

    def __reduce__(self):
        raise RuntimeError("Thread pool is not reducible.")



import inspect
import torch
import torch.nn as nn


def soft_update(target_net: nn.Module,
                source_net: nn.Module,
                update_rate) -> None:
    """
    Soft update target network's parameters.

    Args:
        target_net: Target network to be updated.
        source_net: Source network providing new parameters.
        update_rate: Update rate.

    Returns:
        None
    """
    with torch.no_grad():
        for target_param, param in zip(target_net.parameters(),
                                       source_net.parameters()):
            target_param.data.copy_(
                target_param.data * (1.0 - update_rate) +
                param.data.to(target_param.device) * update_rate
            )


def hard_update(target_net: nn.Module,
                source_net: nn.Module) -> None:
    """
    Hard update (directly copy) target network's parameters.

    Args:
        target_net: Target network to be updated.
        source_net: Source network providing new parameters.
    """

    for target_buffer, buffer in zip(target_net.buffers(),
                                     source_net.buffers()):
        target_buffer.data.copy_(buffer.data)
    for target_param, param in zip(target_net.parameters(),
                                   source_net.parameters()):
        target_param.data.copy_(param.data)


def safe_call(model, *named_args, required_argument=()):
    """
    Call a model and discard unnecessary arguments. safe_call will automatically
    move tensors in named_args to the input device of the model

    Any input tensor in named_args must not be contained inside any container,
    such as list, dict, tuple, etc. Because they will be automatically moved
    to the input device of the specified model.

    Args:
        model: Model to be called, must be a wrapped nn.Module or an instance of
               NeuralNetworkModule.
        named_args: A dictionary of argument, key is argument's name, value is
                    argument's value.
        required_argument: A list/tuple of required arguments' name.

    Returns:
        Whatever returned by your module.
    """
    if (not hasattr(model, "input_device") or
            not hasattr(model, "output_device")):
        raise RuntimeError("Wrap your model of type nn.Module with one of: \n"
                           "1. StaticModuleWrapper from models.models.base \n"
                           "2. DynamicModuleWrapper from models.models.base \n"
                           "Or construct your own module & model with: \n"
                           "NeuralNetworkModule from models.models.base")
    input_device = model.input_device
    args = inspect.getfullargspec(model.forward).args
    args_dict = {}
    if any(arg not in args for arg in required_argument):
        missing = []
        for arg in required_argument:
            if arg not in args:
                missing.append(arg)
        raise RuntimeError("Model missing required argument field(s): {}, "
                           "check your store_observe() function.".format(missing))
    for na in named_args:
        for k, v in na.items():
            if k in args:
                if torch.is_tensor(v):
                    args_dict[k] = v.to(input_device)
                else:
                    args_dict[k] = v
    return model(**args_dict)


def assert_output_is_probs(tensor):
    if tensor.dim() == 2 and \
            torch.all(torch.abs(torch.sum(tensor, dim=1) - 1.0) < 1e-5) and \
            torch.all(tensor > 0):
        return
    else:
        raise RuntimeError("Input tensor is not a probability tensor, it must "
                           "have a dimension of 2, a sum of 1.0 for each row in"
                           " dimension 1, and a positive value for each "
                           "element.")



from typing import Union, Dict, List, Tuple, Callable, Any
from torch.distributions import Categorical
import torch as t
import torch.nn as nn

from machin.frame.buffers.buffer import Transition, Buffer
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework
from .utils import hard_update, soft_update, safe_call


class DQN(TorchFramework):
    """
    DQN framework.
    """

    _is_top = ["qnet", "qnet_target"]
    _is_restorable = ["qnet_target"]

    def __init__(self,
                 qnet: Union[NeuralNetworkModule, nn.Module],
                 qnet_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 mode: str = "double",
                 visualize: bool = False,
                 **__):
        """
        Note:
            DQN is only available for discreet environments.

        Note:
            Dueling DQN is a network structure rather than a framework, so
            it could be applied to all three modes.

            If ``mode = "vanilla"``, implements the simplest online DQN,
            with replay buffer.

            If ``mode = "fixed_target"``, implements DQN with a target network,
            and replay buffer. Described in `this <https://web.stanford.\
edu/class/psych209/Readings/MnihEtAlHassibis15NatureControlDeepRL.pdf>`__ essay.

            If ``mode = "double"``, implements Double DQN described in
            `this <https://arxiv.org/pdf/1509.06461.pdf>`__ essay.
        Note:
            Vanilla DQN only needs one network, so internally, ``qnet``
            is assigned to ``qnet_target``.

        Note:
            In order to implement dueling DQN, you should create two dense
            output layers.

            In your q network::

                    self.fc_adv = nn.Linear(in_features=...,
                                            out_features=num_actions)
                    self.fc_val = nn.Linear(in_features=...,
                                            out_features=1)

            Then in your ``forward()`` method, you should implement output as::

                    adv = self.fc_adv(some_input)
                    val = self.fc_val(some_input).expand(self.batch_sze,
                                                         self.num_actions)
                    return val + adv - adv.mean(1, keepdim=True)

        Note:
            Your optimizer will be called as::

                optimizer(network.parameters(), learning_rate)

            Your lr_scheduler will be called as::

                lr_scheduler(
                    optimizer,
                    *lr_scheduler_args[0],
                    **lr_scheduler_kwargs[0],
                )

            Your criterion will be called as::

                criterion(
                    target_value.view(batch_size, 1),
                    predicted_value.view(batch_size, 1)
                )

        Args:
            qnet: Q network module.
            qnet_target: Target Q network module.
            optimizer: Optimizer used to optimize ``qnet``.
            criterion: Criterion used to evaluate the value loss.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`

            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            reward_func: Reward function used in training.
            mode: one of ``"vanilla", "fixed_target", "double"``.
            visualize: Whether visualize the network flow in the first pass.
        """
        self.batch_size = batch_size
        self.update_rate = update_rate
        self.discount = discount
        self.visualize = visualize

        if mode not in {"vanilla", "fixed_target", "double"}:
            raise RuntimeError("Unknown DQN mode: {}".format(mode))
        self.mode = mode

        self.qnet = qnet
        if self.mode == "vanilla":
            self.qnet_target = qnet
        else:
            self.qnet_target = qnet_target
        self.qnet_optim = optimizer(self.qnet.parameters(),
                                    lr=learning_rate)
        self.replay_buffer = (Buffer(replay_size, replay_device)
                              if replay_buffer is None
                              else replay_buffer)

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.qnet, self.qnet_target)

        if lr_scheduler is not None:
            self.qnet_lr_sch = lr_scheduler(
                self.qnet_optim,
                *lr_scheduler_args[0],
                **lr_scheduler_kwargs[0]
            )

        self.criterion = criterion

        self.reward_func = (DQN.bellman_function
                            if reward_func is None
                            else reward_func)

        super(DQN, self).__init__()

    def act_discreet(self,
                     state: Dict[str, Any],
                     use_target: bool = False,
                     **__):
        """
        Use Q network to produce a discreet action for
        the current state.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Action of shape ``[batch_size, 1]``.
        """
        if use_target:
            result = safe_call(self.qnet_target, state)
        else:
            result = safe_call(self.qnet, state)

        result = t.argmax(result, dim=1).view(-1, 1)
        return result

    def act_discreet_with_noise(self,
                                state: Dict[str, Any],
                                use_target: bool = False,
                                **__):
        """
        Use Q network to produce a noisy discreet action for
        the current state.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Noisy action of shape ``[batch_size, 1]``.
        """
        if use_target:
            result = safe_call(self.qnet_target, state)
        else:
            result = safe_call(self.qnet, state)

        result = t.softmax(result, dim=1)
        dist = Categorical(result)
        batch_size = result.shape[0]
        return dist.sample([batch_size, 1])

    def criticize(self,
                  state: Dict[str, Any],
                  use_target: bool = False,
                  **__):
        """
        Use Q network to evaluate current value.

        Args:
            state: Current state.
            use_target: Whether to use the target network.

        Returns:
            Q value of shape ``[batch_size, action_dim]``.
        """
        if use_target:
            return safe_call(self.qnet_target, state)
        else:
            return safe_call(self.qnet, state)

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.
        """
        self.replay_buffer.append(transition, required_attrs=(
            "state", "action", "reward", "next_state", "terminal"
        ))

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        for trans in episode:
            self.replay_buffer.append(trans, required_attrs=(
                "state", "action", "reward", "next_state", "terminal"
            ))

    def update(self,
               update_value=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether update the Q network.
            update_target: Whether update targets.
            concatenate_samples: Whether concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """
        batch_size, (state, action, reward, next_state, terminal, *others) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        if self.mode == "vanilla":
            # Vanilla DQN, directly optimize q network.
            # target network is the same as the main network
            q_value = self.criticize(state)
            action_value = q_value.gather(dim=1,
                                          index=action["action"]
                                          .to(q_value.device))
            target_next_q_value = t.max(self.criticize(next_state), dim=1)[0]\
                                   .unsqueeze(1).detach()
            y_i = self.reward_func(reward, self.discount, target_next_q_value,
                                   terminal, *others)
            value_loss = self.criterion(action_value,
                                        y_i.to(action_value.device))

            if self.visualize:
                self.visualize_model(value_loss, "qnet")

            if update_value:
                self.qnet.zero_grad()
                value_loss.backward()
                self.qnet_optim.step()

        elif self.mode == "fixed_target":
            # Fixed target DQN, which estimate next value by using the
            # target Q network. Similar to the idea of DDPG.
            q_value = self.criticize(state)
            action_value = q_value.gather(dim=1,
                                          index=action["action"]
                                          .to(q_value.device))
            target_next_q_value = t.max(self.criticize(next_state, True),
                                        dim=1)[0].unsqueeze(1).detach()

            y_i = self.reward_func(reward, self.discount, target_next_q_value,
                                   terminal, *others)
            value_loss = self.criterion(action_value,
                                        y_i.to(action_value.device))

            if self.visualize:
                self.visualize_model(value_loss, "qnet")

            if update_value:
                self.qnet.zero_grad()
                value_loss.backward()
                self.qnet_optim.step()

            # Update target Q network
            if update_target:
                soft_update(self.qnet_target, self.qnet, self.update_rate)

        elif self.mode == "double":
            # Double DQN. DDQN also use the target network to estimate the next
            # value, but instead of selecting the maximum Q(s,a), it uses
            # the online DQN network to select an action and return Q(s,a'), to
            # reduce the over estimation.
            q_value = self.criticize(state)
            action_value = q_value.gather(dim=1,
                                          index=action["action"]
                                          .to(q_value.device))

            with t.no_grad():
                next_q_value = self.criticize(next_state)
                target_next_q_value = self.criticize(next_state, True)
                target_next_q_value = target_next_q_value.gather(
                    dim=1, index=t.max(next_q_value, dim=1)[1].unsqueeze(1))

            y_i = self.reward_func(reward, self.discount, target_next_q_value,
                                   terminal, *others)
            value_loss = self.criterion(action_value,
                                        y_i.to(action_value.device))

            if self.visualize:
                self.visualize_model(value_loss, "qnet")

            if update_value:
                self.qnet.zero_grad()
                value_loss.backward()
                self.qnet_optim.step()

            # Update target Q network
            if update_target:
                soft_update(self.qnet_target, self.qnet, self.update_rate)

        else:
            raise RuntimeError("Unknown DQN mode: {}".format(self.mode))

        # use .item() to prevent memory leakage
        return value_loss.item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "qnet_lr_sch"):
            self.qnet_lr_sch.step()

    def load(self, model_dir, network_map=None, version=-1):
        # DOC INHERITED
        super(DQN, self).load(model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.qnet, self.qnet_target)

    @staticmethod
    def bellman_function(reward, discount, next_value, terminal, *_):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * (1 - terminal) * next_value



from typing import Any
from torch.multiprocessing.reductions import ForkingPickler
import dill
import torch as t


def mark_static_module(module: Any):
    """
    Some modules are **static**, which means they are stateless
    and will remain the same whether you import it in process A
    or process B.

    If your module contains reference to functions, objects
    or anything inside a CDLL (usually the reference is a
    pointer), it is not pickable by dill, and will cause
    nasty errors, however, by marking this module as "Static",
    dill will recognize this module as a builtin module and
    not saving the states of this module, dill will only save
    a reference to it in this situation.

    Args:
        module: Some module which imports CDLLs by hand and
            not using pybind11.
    """
    del module.__file__


def dump_tensor(tensor: t.Tensor, reduce_as_reference=False):
    """
    Convert tensor to bytes. Works for cpu and gpu tensors.

    Warning:
        Till pytorch 1.5.0, there is a bug for referenced gpu tensors,
        which would require users to keep shared gpu tensors during
        the whole process life and not reassigning / deleting them,
        however, you may refill them with different values.

        See `here <https://github.com/pytorch/pytorch/issues/39541>`_

    Args:
        tensor: Some tensor.
        reduce_as_reference: Whether to dump the tensor as a reference,
            useful in same-node cross-process tensor transmission.

    Returns:
        Bytes.
    """
    if reduce_as_reference:
        # Only works on the same node, where tensor transmission
        # is needed between processes.
        # Since the registered reduction functions also return
        # constructor function (in torch/multiprocessing/reductions.py)
        # we can use dill to directly loads tensors.
        return ForkingPickler.dumps(tensor)
    else:
        # Pytorch implemented __reduce__ in the _StorageBase class
        # Storage is the base of all types of storage, and the
        # container of all types of tensors.
        # Internally it uses torch.save to convert the tensor to
        # raw bytes.
        tensor.share_memory_()
        return dill.dumps(tensor)



from typing import Union, Dict, List, Tuple, Callable, Any
import torch as t
import torch.nn as nn
import numpy as np

from machin.frame.buffers.buffer import Transition, Buffer
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework
from .utils import hard_update, soft_update, safe_call


class SAC(TorchFramework):
    """
    SAC framework.
    """

    _is_top = ["actor", "critic", "critic2", "critic_target", "critic2_target"]
    _is_restorable = ["actor", "critic_target", "critic2_target"]

    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 critic2: Union[NeuralNetworkModule, nn.Module],
                 critic2_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 target_entropy: float = None,
                 initial_entropy_alpha: float = 1.0,
                 gradient_max: float = np.inf,
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 action_trans_func: Callable = None,
                 visualize: bool = False,
                 **__):
        """
        See Also:
            :class:`.A2C`
            :class:`.DDPG`

        Important:
            when given a state, and an optional, action actor must
            at least return two values, similar to the actor structure
            described in :class:`.A2C`. You may use the same actor model
            here.

            However, Compared to other actor-critic methods, SAC embeds the
            entropy term into its reward function directly, rather than adding
            the entropy term to actor's loss function. Therefore, we do not need
            the actor network to output a entropy as its third returned result.

            The SAC algorithm uses Q network as critics, so please reference
            :class:`.DDPG` for the requirements and the definition of
            ``action_trans_func``.

        Args:
            actor: Actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            critic2: The second critic network module.
            critic2_target: The second target critic network module.
            optimizer: Optimizer used to optimize ``actor``, ``critic`` and
                ``critic2``.
            criterion: Criterion used to evaluate the value loss.
            *_:
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            target_entropy: Target entropy weight :math:`\\alpha` used in
                the SAC soft value function:
                :math:`V_{soft}(s_t) = \\mathbb{E}_{q_t\\sim\\pi}[\
                                        Q_{soft}(s_t,a_t) - \
                                        \\alpha log\\pi(a_t|s_t)]`
            initial_entropy_alpha: Initial entropy weight :math:`\\alpha`
            gradient_max: Maximum gradient.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            reward_func: Reward function used in training.
            action_trans_func: Action transform function, used to transform
                the raw output of your actor, by default it is:
                ``lambda act: {"action": act}``
            visualize: Whether visualize the network flow in the first pass.
            **__:
        """
        self.batch_size = batch_size
        self.update_rate = update_rate
        self.discount = discount
        self.visualize = visualize
        self.entropy_alpha = t.tensor(initial_entropy_alpha,
                                      requires_grad=True).view(1)
        self.grad_max = gradient_max
        self.target_entropy = target_entropy

        self.actor = actor
        self.critic = critic
        self.critic_target = critic_target
        self.critic2 = critic2
        self.critic2_target = critic2_target
        self.actor_optim = optimizer(self.actor.parameters(),
                                     lr=learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(),
                                      lr=learning_rate)
        self.critic2_optim = optimizer(self.critic2.parameters(),
                                       lr=learning_rate)
        self.alpha_optim = optimizer([self.entropy_alpha],
                                     lr=learning_rate)
        self.replay_buffer = (Buffer(replay_size, replay_device)
                              if replay_buffer is None
                              else replay_buffer)

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.critic, self.critic_target)
            hard_update(self.critic2, self.critic2_target)

        if lr_scheduler is not None:
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim,
                *lr_scheduler_args[0],
                **lr_scheduler_kwargs[0],
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim,
                *lr_scheduler_args[1],
                **lr_scheduler_kwargs[1]
            )
            self.critic2_lr_sch = lr_scheduler(
                self.critic2_optim,
                *lr_scheduler_args[1],
                **lr_scheduler_kwargs[1]
            )

        self.criterion = criterion

        self.reward_func = (SAC.bellman_function
                            if reward_func is None
                            else reward_func)
        self.action_trans_func = (SAC.action_transform_function
                                  if action_trans_func is None
                                  else action_trans_func)

        super(SAC, self).__init__()

    def act(self, state: Dict[str, Any], **__):
        """
        Use actor network to produce an action for the current state.

        Returns:
            Anything produced by actor.
        """
        return safe_call(self.actor, state)

    def criticize(self,
                  state: Dict[str, Any],
                  action: Dict[str, Any],
                  use_target: bool = False,
                  **__):
        """
        Use critic network to evaluate current value.

        Args:
            state: Current state.
            action: Current action.
            use_target: Whether to use the target network.

        Returns:
            Value evaluated by critic.
        """
        if use_target:
            return safe_call(self.critic_target, state, action)
        else:
            return safe_call(self.critic, state, action)

    def criticize2(self,
                   state: Dict[str, Any],
                   action: Dict[str, Any],
                   use_target: bool = False,
                   **__):
        """
        Use the second critic network to evaluate current value.

        Returns:
            Value evaluated by critic.
        """
        if use_target:
            return safe_call(self.critic2_target, state, action)
        else:
            return safe_call(self.critic2, state, action)

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.
        """
        self.replay_buffer.append(transition, required_attrs=(
            "state", "action", "next_state", "reward", "terminal"
        ))

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        for trans in episode:
            self.replay_buffer.append(trans, required_attrs=(
                "state", "action", "next_state", "reward", "terminal"
            ))

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               update_entropy_alpha=True,
               concatenate_samples=True,
               **__):
        """
        Update network weights by sampling from replay buffer.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.
            update_target: Whether to update targets.
            update_entropy_alpha: Whether to update :math:`alpha` of entropy.
            concatenate_samples: Whether to concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """
        batch_size, (state, action, reward, next_state, terminal, *others) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        # Update critic network first
        with t.no_grad():
            next_action, next_action_log_prob, _ = self.actor(next_state)
            next_action = self.action_trans_func(next_action, next_state,
                                                 *others)
            next_value = self.criticize(next_state, next_action, True)
            next_value2 = self.criticize2(next_state, next_action, True)
            next_value = t.min(next_value, next_value2)
            next_value = (next_value.view(batch_size, -1) -
                          self.entropy_alpha
                          * next_action_log_prob.view(batch_size, -1))
            y_i = self.reward_func(reward, self.discount, next_value,
                                   terminal, *others)

        cur_value = self.criticize(state, action)
        cur_value2 = self.criticize2(state, action)
        value_loss = self.criterion(cur_value, y_i.to(cur_value.device))
        value_loss2 = self.criterion(cur_value2, y_i.to(cur_value.device))

        if self.visualize:
            self.visualize_model(value_loss, "critic")

        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            self.critic_optim.step()

            self.critic2.zero_grad()
            value_loss2.backward()
            self.critic2_optim.step()

        # Update actor network
        cur_action, cur_action_log_prob, *_ = self.actor(next_state)
        cur_action = self.action_trans_func(cur_action, state, *others)
        act_value = self.criticize(state, cur_action)
        act_value2 = self.criticize2(state, cur_action)
        act_value = t.min(act_value, act_value2)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = ((self.entropy_alpha * cur_action_log_prob) -
                           act_value).mean()

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor")

        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            self.actor_optim.step()

        # Update target networks
        if update_target:
            soft_update(self.critic_target, self.critic, self.update_rate)

        if update_entropy_alpha:
            alpha_loss = -(t.log(self.entropy_alpha) *
                           (cur_action_log_prob + self.target_entropy).detach()
                           ).mean()
            self.alpha_optim.zero_grad()
            alpha_loss.backward()
            self.alpha_optim.step()

        # use .item() to prevent memory leakage
        return -act_policy_loss.item(), value_loss.item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()

    def load(self, model_dir, network_map=None, version=-1):
        # DOC INHERITED
        super(SAC, self).load(model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.critic, self.critic_target)
            hard_update(self.critic, self.critic2_target)

    @staticmethod
    def action_transform_function(raw_output_action, *_):
        return {"action": raw_output_action}

    @staticmethod
    def bellman_function(reward, discount, next_value, terminal, *_):
        next_value = next_value.to(reward.device)
        terminal = terminal.to(reward.device)
        return reward + discount * (1 - terminal) * next_value



from typing import Union, Dict, List, Tuple, Callable, Any
import numpy as np
import torch as t
import torch.nn as nn

from machin.frame.buffers.buffer_d import Transition, DistributedBuffer
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework
from .utils import safe_call

from machin.parallel.server import PushPullModelServer
from machin.parallel.distributed import RpcGroup


def _slice(attr_dict, index, dim=1):
    # Used to split a specific slice from a transition main attribute
    new_dict = {}
    for k, v in attr_dict.items():
        new_dict[k] = v.select(dim, index)
    return new_dict


def _make_tensor_from_batch(batch: List[Any], device, concatenate):
    """
    Used to convert a whole episode to a single "Transition".

    Args:
        batch:
        device: Device to move tensors to.
        concatenate: Whether to perform concatenation or not, only True
            for major and sub attributes in ``Transition``.

    Returns:
        A tensor if ``concatenate`` is ``True``, otherwise the original List.
    """
    if len(batch) == 0:
        return None
    if concatenate:
        item = batch[0]
        batch_size = len(batch)
        if t.is_tensor(item):
            batch = [it.to(device) for it in batch]
            if item.dim() >= 3:
                # ignore size difference in dimension 1 and pad with 0
                # if there if no size difference, result should be the
                # same as cat
                pad_length = max([it.shape[1] for it in batch])
                remain_shape = [item.shape[2:]]
                result = t.zeros([batch_size, pad_length] + remain_shape,
                                 dtype=item.dtype, device=device)
                for it, idx in zip(batch, range(batch_size)):
                    result[idx, :it.shape[1]] = it
                return result
            else:
                return t.cat(batch, dim=0).to(device)
        else:
            return t.tensor(batch, device=device).view(batch_size, -1)
    else:
        return batch


class IMPALA(TorchFramework):
    """
    Massively parallel IMPALA framework.
    """

    _is_top = ["actor", "critic"]
    _is_restorable = ["actor", "critic"]

    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 worker_group: RpcGroup,
                 trainer_group: RpcGroup,
                 *_,
                 pull_function: Callable = None,
                 push_function: Callable = None,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 learning_rate=0.001,
                 isw_clip_c: float = 0.8,
                 isw_clip_rho: float = 0.9,
                 entropy_weight: float = None,
                 value_weight: float = 0.5,
                 gradient_max: float = np.inf,
                 discount: float = 0.99,
                 batch_size: int = 5,
                 replay_size: int = 500,
                 visualize: bool = False,
                 **__):
        """
        TODO: test IMPAlA and add more explanations in document.
        TODO: add visualization in update.

        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            worker_group: Rpc group of roll out workers.
            trainer_group: Rpc group of model trainers.
            pull_function: User defined function used to pull the newest model.
            push_function: User defined function used to push the newest model.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            isw_clip_c: :math:`c` used in importance weight clipping.
            isw_clip_rho:
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            gradient_max: Maximum gradient.
            discount: :math:`\\gamma` used in the bellman function.
            batch_size: Batch size used during training.
            replay_size: Size of the local replay buffer.
            visualize: Whether visualize the network flow in the first pass.
        """
        self.batch_size = batch_size
        self.discount = discount
        self.value_weight = value_weight
        self.entropy_weight = entropy_weight
        self.grad_max = gradient_max
        self.isw_clip_c = isw_clip_c
        self.isw_clip_rho = isw_clip_rho
        self.visualize = visualize

        self.worker_group = worker_group
        self.trainer_group = trainer_group

        self.actor = actor
        self.critic = critic
        self.actor_optim = optimizer(self.actor.parameters(),
                                     lr=learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(),
                                      lr=learning_rate)
        self.replay_buffer = DistributedBuffer(buffer_size=replay_size,
                                               buffer_group=worker_group)

        if push_function is None or pull_function is None:
            self.pp = PushPullModelServer(trainer_group)
            self.pull_function = self.pp.pull
            self.push_function = self.pp.push
        else:
            self.pull_function = pull_function
            self.push_function = push_function

        if lr_scheduler is not None:
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim,
                *lr_scheduler_args[0],
                **lr_scheduler_kwargs[0],
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim,
                *lr_scheduler_args[1],
                **lr_scheduler_kwargs[1]
            )

        self.criterion = criterion

        super(IMPALA, self).__init__()

    def act(self, state: Dict[str, Any], *_, **__):
        """
        Use actor network to give a policy to the current state.

        Returns:
            Anything produced by actor.
        """
        return safe_call(self.actor, state)

    def eval_act(self,
                 state: Dict[str, Any],
                 action: Dict[str, Any],
                 *_, **__):
        """
        Use actor network to evaluate the log-likelihood of a given
        action in the current state.

        Returns:
            Anything produced by actor.
        """
        self.pull_function(self.actor, "actor")
        return safe_call(self.actor, state, action)

    def criticize(self, state: Dict[str, Any], *_, **__):
        """
        Use critic network to evaluate current value.

        Returns:
            Value evaluated by critic.
        """
        self.pull_function(self.critic, "critic")
        return safe_call(self.critic, state)

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Warning:
            Not supported in IMPALA due to v-trace requirements.
        """
        raise NotImplementedError

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        if not isinstance(episode[0], Transition):
            episode = [Transition(**trans) for trans in episode]

        cc_episode = {}
        # In order to compute v-trace, we must reshape the whole
        # episode to make it look like a single Transition, because
        # v-trace need to see all future rewards.
        for k, v in episode[0].items():
            if k in ("state", "action", "next_state"):
                tmp_dict = {}
                for sub_k in v.keys():
                    tmp_dict[sub_k] = _make_tensor_from_batch(
                        [item[k][sub_k] for item in episode],
                        self.replay_buffer.buffer_device, True
                    ).unsqueeze(0)
                cc_episode[k] = tmp_dict
            elif k in ("reward", "terminal"):
                cc_episode[k] = _make_tensor_from_batch(
                    [item[k] for item in episode],
                    self.replay_buffer.buffer_device, True
                ).unsqueeze(0)
            else:
                cc_episode[k] = _make_tensor_from_batch(
                    [item[k] for item in episode],
                    self.replay_buffer.buffer_device, False
                )

        self.replay_buffer.append(cc_episode)

    def update(self,
               update_value=True,
               update_policy=True,
               **__):
        """
        Update network weights by sampling from replay buffer.

        Note:
            Will always concatenate samples.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.

        Returns:
            mean value of estimated policy value, value loss
        """
        # sample a batch
        batch_size, (state, action, reward, next_state,
                     terminal, action_log_prob) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate=True,
                                            device="cpu",
                                            sample_attrs=[
                                                "state", "action", "reward",
                                                "next_state", "terminal",
                                                "action_log_prob"],
                                            additional_concat_attrs=[
                                                "action_log_prob"
                                            ])
        step_len = reward.shape[1]
        term = t.zeros([batch_size], dtype=t.bool)
        values = []
        log_probs = []
        entropies = []
        c = []
        rho = []
        clip_c = t.full(action_log_prob.shape, self.isw_clip_c)
        clip_rho = t.full(action_log_prob.shape, self.isw_clip_rho)

        # pre process
        for step in range(step_len):
            term = term | terminal[:, step]
            terminal[:, step] = term
            value = self.criticize(_slice(state, step)).to("cpu")

            if self.entropy_weight is not None:
                _, new_action_log_prob, new_action_entropy, *_ = \
                    self.eval_act(_slice(state, step), _slice(action, step))
                entropies.append(new_action_entropy)
            else:
                _, new_action_log_prob, *_ = \
                    self.eval_act(_slice(state, step), _slice(action, step))

            new_action_log_prob = new_action_log_prob.to("cpu")
            # importance sampling weight
            is_weight = t.exp(new_action_log_prob - action_log_prob)

            values.append(value)
            log_probs.append(new_action_log_prob)
            c.append(t.min(clip_c, is_weight))
            rho.append(t.min(clip_rho, is_weight))

        act_policy_loss = t.zeros([1])

        # v-trace, ``vs`` is v-trace target
        vs = t.zeros([step_len, batch_size, 1], device="cpu")
        for rev_step in reversed(range(step_len - 1)):
            # terminal is not in the original essay? Maybe it is useful?
            delta_v = rho[rev_step] * (reward[:, rev_step]
                                       * (1 - terminal[:, rev_step]) +
                                       self.discount * values[rev_step + 1] -
                                       values[rev_step])
            advantage = rho[rev_step] * (reward[:, rev_step]
                                         * (1 - terminal[:, rev_step]) +
                                         self.discount * vs[rev_step + 1] -
                                         values[rev_step])
            vs[rev_step] = (values[rev_step] +
                            delta_v +
                            self.discount
                            * c[rev_step]
                            * (vs[rev_step + 1] - values[rev_step + 1]))

            act_policy_loss += log_probs[rev_step] * advantage.detach()

        act_policy_loss += self.entropy_weight * t.sum(t.stack(entropies))
        act_policy_loss = act_policy_loss.sum()

        value_loss = self.criterion(t.stack(values).to(vs.device),
                                    vs.transpose(0, 1).detach())

        # Update actor network
        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            nn.utils.clip_grad_norm_(
                self.actor.parameters(), self.grad_max
            )
            self.actor_optim.step()
            self.push_function(self.actor, "actor")

        # Update critic network
        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            nn.utils.clip_grad_norm_(
                self.critic.parameters(), self.grad_max
            )
            self.critic_optim.step()
            self.push_function(self.critic, "critic")

        return -act_policy_loss.item(), value_loss.item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()



from time import sleep
from random import shuffle
from torch.distributed import rpc
from abc import ABC, abstractmethod
from typing import List, Tuple
from threading import Thread, Condition, Lock


class RoleDispatcherBase(ABC):
    """
    Base role dispatcher.
    """
    @abstractmethod
    def __init__(self,
                 rank: int,
                 world_size: int,
                 roles: List[str],
                 role_counts: List[int]):
        """
        Args:
            rank: Absolute current process rank.
            world_size: Size of the whole world.
            roles: A list of role names.
            role_counts: A list of numbers of each role.
        """
        def f(*_):
            pass
        # Prevent pylint from complaining
        f(rank, world_size, roles, role_counts)
        self.has_failure_cond = Condition(Lock())
        self.role_update_cond = Condition(Lock())

    @abstractmethod
    def start(self) -> None:
        """
        Start the role dispatcher.
        """
        pass

    @abstractmethod
    def get_roles(self) -> List[Tuple[str, int]]:
        """
        Returns:
            All roles.
        """
        pass

    @abstractmethod
    def get_rank(self, role: Tuple[str, int]) -> int:
        """
        Get the assigned process rank of the specified role.

        Args:
            role: A role tuple like ("some_role", 12), the number indicates
                the role index in its role category.

        Returns:
            Absolute assigned process rank.
        """
        pass

    def get_role_update_cond(self):
        """
        Returns:
            the role update conditional.
        """
        return self.role_update_cond

    @abstractmethod
    def notify_failure(self, role: Tuple[str, int]) -> None:
        """
        Upper RPC layer will notify role dispatcher the failure of a role,
        dispatcher may reassign this role to other processes immediately,
        or wait for a while, RPC will retry until success if retrying is
        permitted by the user.

        Args:
            role: A role tuple like ("some_role", 12), the number indicates
                the role index in its role category.
        """
        pass


class RoleDispatcherSimple(RoleDispatcherBase):
    """
    The simplest static role dispatcher. Requires process number
    be larger or equal to the role number.
    """
    def __init__(self, rank, world_size, roles, role_counts):
        # DOC INHERITED
        idx = 0
        role = 0
        for role_count in role_counts:
            idx += role_count
            if idx > rank:
                self.roles = [roles[role]]
                break
            role += 1
        else:
            # Extra processes are ignored.
            self.role = []

        # Calculate prefixed role offset so we can lookup
        # the absolute process rank by computing ``role_index + offset``
        self.prefixed_role_count = {}
        count_sum = 0
        for role, role_count in zip(roles, role_counts):
            self.prefixed_role_count[role] = count_sum
            count_sum += role_count

        super(RoleDispatcherSimple, self).__init__(
            rank, world_size, roles, role_counts
        )

    def start(self):
        # DOC INHERITED
        pass

    def get_roles(self):
        # DOC INHERITED
        return self.roles

    def get_rank(self, role):
        # DOC INHERITED
        return self.prefixed_role_count[role[0]] + role[1]

    def notify_failure(self, role: Tuple[str, int]) -> None:
        # DOC INHERITED
        pass


class RoleDispatcherElection(RoleDispatcherBase):
    """
    Role dispatcher based on election group. The election group
    leader wil serve as the dispatcher and allocate roles to processes.
    """
    def __init__(self, rank, world_size, role_names, role_nums,
                 election_group):
        self.elect_group = election_group

        self.lead_thread = Thread(target=self._task_lead_group)
        self.deal_fail_thread = Thread(target=self._task_deal_failure)
        # This lock is used to prevent the lead thread and deal_fail
        # thread both execute the reassign task.
        self.assign_lock = Lock()

        self.dispatched_roles = []
        self.roles = []
        self.ranks = {}
        self.all_roles = set()
        for role, role_count in zip(role_names, role_nums):
            for idx in range(role):
                self.all_roles.add((role, idx))

        super(RoleDispatcherElection, self).__init__(
            rank, world_size, role_names, role_nums
        )

    def start(self):
        # DOC INHERITED
        self.lead_thread.start()
        self.deal_fail_thread.start()
        # Elect thread must be started before the elect group to
        # let it wait on the leader change signal.
        self.elect_group.start()

    def get_roles(self):
        # DOC INHERITED
        return self.roles

    def get_rank(self, role):
        # DOC INHERITED
        return self.ranks[role]

    def notify_failure(self, role):
        # DOC INHERITED
        self.has_failure_cond.notify_all()

    def _task_lead_group(self):
        # Wait for leader to change
        self.elect_group.get_leader_change_cond().wait()
        self.assign_lock.acquire()
        while (self.elect_group.is_leader()
               and not self._assign_roles()
               and not self._update_ranks()):
            sleep(0.1)
        self.assign_lock.release()

    def _task_deal_failure(self):
        # Deal with failure
        self.has_failure_cond.wait()
        self.assign_lock.acquire()
        while (self.elect_group.is_leader()
               and not self._assign_roles()
               and not self._update_ranks()):
            sleep(0.1)
        self.assign_lock.release()

    def _assign_roles(self):
        future = []
        assigned_roles = set()
        available_members = []
        unassigned_members = []
        # Get assigned roles from all processes, also test their aliveness.
        for member in self.elect_group.get_members():
            # TODO: add timeout
            future.append((
                member,
                rpc.rpc_async(str(member),
                              RoleDispatcherElection._rpc_respond_roles)
            ))
        for member, fut in future:
            try:
                member_roles = fut.wait()
                assigned_roles.update(member_roles)
                if len(member_roles) == 0:
                    unassigned_members.append(member)
                available_members.append(member)
            except RuntimeError:
                # timeout, peer reset (disconnect) etc.
                continue

        # Assign unassigned roles.
        unassigned_roles = list(self.all_roles - assigned_roles)
        assign_map = self._plan_assign_roles(unassigned_roles,
                                             unassigned_members,
                                             available_members)

        if assign_map is None:
            # Failed to assign, restart assign process.
            return False

        future.clear()
        success = []

        # Dispatch unassigned roles to all processes.
        for member, roles in assign_map.items():
            # TODO: add timeout
            future.append(
                rpc.rpc_async(
                    str(member),
                    RoleDispatcherElection._rpc_dispatch_roles,
                    args=(roles,)
                )
            )
        for fut in future:
            try:
                fut.wait()
                success.append(True)
            except RuntimeError:
                # timeout, peer reset (disconnect) etc.
                # Failed to assign, restart assign process.
                success.append(False)
        return all(success)

    @staticmethod
    def _plan_assign_roles(roles, unassigned_members, available_members):
        # Decide how to assign unassigned roles.
        shuffle(available_members)
        result = {member: [] for member in available_members}

        # assign roles to free members first
        idx = 0
        if len(available_members) == 0:
            return None
        for role in roles:
            if len(unassigned_members) > 0:
                result[unassigned_members[0]].append(role)
                unassigned_members.pop(0)
            else:
                # Assign remaining roles in a round robin fashion.
                result[available_members[idx]].append(role)
                idx = (idx + 1) % len(available_members)
        return result

    def _update_ranks(self):
        future = []
        # Get assigned roles from all processes, also test their aliveness.
        for member in self.elect_group.get_members():
            # TODO: add timeout
            future.append((
                str(member),
                rpc.rpc_async(member, RoleDispatcherElection._rpc_respond_roles)
            ))

        for member, fut in future:
            try:
                member_roles = fut.wait()
                for member_role in member_roles:
                    self.ranks[member_role] = member
            except RuntimeError:
                # timeout, peer reset (disconnect) etc.
                return False

        future.clear()
        # Push role-rank route table to all processes.
        for member in self.elect_group.get_members():
            # TODO: add timeout
            future.append((
                str(member),
                rpc.rpc_async(member, RoleDispatcherElection._rpc_set_ranks,
                              args=(self.ranks,))
            ))
        for member, fut in future:
            try:
                fut.wait()
            except RuntimeError:
                # timeout, peer reset (disconnect) etc.
                return False
        return True

    def _rpc_respond_roles(self):
        return self.roles

    def _rpc_dispatch_roles(self, roles):
        self.roles += roles
        self.role_update_cond.notify_all()

    def _rpc_set_ranks(self, new_ranks):
        self.ranks = new_ranks



from typing import Union, Dict, Iterable, Any, NewType
from itertools import chain
from copy import deepcopy
import torch as t
import numpy as np

Scalar = NewType("Scalar", Union[int, float, bool])


class TransitionBase(object):
    """
    Base class for all transitions
    """
    def __init__(self,
                 major_attr: Iterable[str],
                 sub_attr: Iterable[str],
                 custom_attr: Iterable[str],
                 major_data: Iterable[Dict[str, t.Tensor]],
                 sub_data: Iterable[Union[Scalar, t.Tensor]],
                 custom_data: Iterable[Any]):
        """
        Note:
            Major attributes store things like state, action, next_states, etc.
            They are usually **concatenated by their dictionary keys** during
            sampling, and passed as keyword arguments to actors, critics, etc.

            Sub attributes store things like terminal states, reward, etc.
            They are usually **concatenated directly** during sampling, and used
            in different algorithms.

            Custom attributes store not concatenatable values, usually user
            specified states, used in models or as special arguments in
            different algorithms. They will be collected together as a list
            during sampling, **no further concatenation is performed**.

        Args:
            major_attr: A list of major attribute names.
            sub_attr: A list of sub attribute names.
            custom_attr: A list of custom attribute names.
            major_data: Data of major attributes.
            sub_data: Data of sub attributes.
            custom_data: Data of custom attributes.
        """
        self._major_attr = list(major_attr)
        self._sub_attr = list(sub_attr)
        self._custom_attr = list(custom_attr)
        self._keys = self._major_attr + self._sub_attr + self._custom_attr
        self._length = self._major_attr + self._sub_attr + self._custom_attr

        for attr, data in zip(chain(major_attr, sub_attr, custom_attr),
                              chain(major_data, sub_data, custom_data)):
            object.__setattr__(self, attr, data)
        self._copy_and_detach()
        self._check_validity()

    def __len__(self):
        return self._length

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        object.__setattr__(key, value)

    def __setattr__(self, key, value):
        object.__setattr__(key, value)

    @property
    def major_attr(self):
        return self._major_attr

    @property
    def sub_attr(self):
        return self._sub_attr

    @property
    def custom_attr(self):
        return self._custom_attr

    def keys(self):
        """
        Returns:
            All attribute names in current transition object.
        """
        return self._keys

    def items(self):
        """
        Returns:
            All attribute values in current transition object.
        """
        for k in self._keys:
            yield k, getattr(self, k)

    def has_keys(self, keys: Iterable[str]):
        """
        Args:
            keys: A list of keys

        Returns:
            A bool indicating whether current transition object
            contains all specified keys.
        """
        return all([k in self._keys for k in keys])

    def to(self, device: Union[str, t.device]):
        """
        Move current transition object to another device. will be
        a no-op if it already locates on that device.

        Args:
            device: A valid pytorch device.

        Returns:
            Self.
        """
        for ma in self._major_attr:
            ma_data = getattr(self, ma)
            for k, v in ma_data.items():
                ma_data[k] = v.to(device)
        for sa in self._sub_attr:
            sa_data = getattr(self, sa)
            if t.is_tensor(sa_data):
                object.__setattr__(self, sa, sa_data.to(device))
        return self

    def _copy_and_detach(self):
        """
        Copy and detach all tensors in major attributes and sub attributes,
        as well as deep-copying various custom attributes.

        Returns:
            Self.
        """
        for ma in self._major_attr:
            ma_data = getattr(self, ma)
            for k, v in ma_data.items():
                ma_data[k] = v.clone().detach()
        for sa in self._sub_attr:
            sa_data = getattr(self, sa)
            if t.is_tensor(sa_data):
                object.__setattr__(self, sa, sa_data.clone().detach())
        for ca in self._custom_attr:
            ca_data = getattr(self, ca)
            object.__setattr__(self, ca, deepcopy(ca_data))
        return self

    def _check_validity(self):
        """
        Check validity of current transition object, will check batch size,
        and major attributes' data, sub attributes' data.

        Raises:
            ``ValueError`` if anything is invalid.
        """
        batch_size = None
        for ma in self._major_attr:
            ma_data = getattr(self, ma)
            for k, v in ma_data.items():
                if not t.is_tensor(v) or v.dim() < 1:
                    raise ValueError("Key {} of transition major attribute "
                                     "{} is a invalid tensor".format(k, ma))
                if batch_size is None:
                    batch_size = v.shape[0]
                else:
                    if batch_size != v.shape[0]:
                        raise ValueError("Key {} of transition major attribute "
                                         "{} has invalid batch size {}."
                                         .format(k, ma, v.shape[0]))
        for sa in self._sub_attr:
            sa_data = getattr(self, sa)
            if np.isscalar(sa_data):
                # will return true for inbuilt scalar types
                # like int, bool, float
                if batch_size != 1:
                    raise ValueError("Transition sub attribute "
                                     "{} is a scalar, but batch size is {}."
                                     .format(sa, batch_size))
            elif t.is_tensor(sa_data):
                if sa_data.dim() < 1:
                    raise ValueError("Transition sub attribute "
                                     "{} is a invalid tensor.")
                elif sa_data.shape[0] != batch_size:
                    raise ValueError("Transition sub attribute "
                                     "{} has invalid batch size {}."
                                     .format(sa, sa_data.shape[0]))
            else:
                raise ValueError("Transition sub attribute {} has invalid "
                                 "value {}, requires scalar or tensor."
                                 .format(sa, sa_data))


class Transition(TransitionBase):
    """
    The default Transition class.

    Have three main attributes: ``state``, ``action`` and ``next_state``.

    Have two sub attributes: ``reward`` and ``terminal``.

    Store one transition step of one agent.
    """
    state = None  # type: Dict[str, t.Tensor]
    action = None  # type: Dict[str, t.Tensor]
    next_state = None  # type: Dict[str, t.Tensor]
    reward = None  # type: Union[float, t.Tensor]
    terminal = None  # type: bool

    def __init__(self,
                 state: Dict[str, t.Tensor],
                 action: Dict[str, t.Tensor],
                 next_state: Dict[str, t.Tensor],
                 reward: Union[float, t.Tensor],
                 terminal: bool,
                 **kwargs):
        """
        Args:
            state: Previous observed state.
            action: Action of agent.
            next_state: Next observed state.
            reward: Reward of agent.
            terminal: Whether environment has reached terminal state.
            **kwargs: Custom attributes.

        Note:
            You should not store any tensor inside ``**kwargs`` as they will
            not be moved to the sample output device.
        """
        super(Transition, self).__init__(
            major_attr=["state", "action", "next_state"],
            sub_attr=["reward", "terminal"],
            custom_attr=kwargs.keys(),
            major_data=[state, action, next_state],
            sub_data=[reward, terminal],
            custom_data=kwargs.values()
        )



from typing import Union, Dict, List, Any, Callable
from threading import Lock
from ..transition import Transition
from .buffer import Buffer
from machin.parallel.distributed import RpcGroup
import torch as t
import numpy as np
import itertools as it


def _round_up(num):
    return int(np.ceil(num))


class DistributedBuffer(Buffer):
    def __init__(self, buffer_size: int, buffer_group: RpcGroup, *_, **__):
        """
        Create a distributed replay buffer instance.

        To avoid issues caused by tensor device difference, all transition
        objects are stored in device "cpu".

        Distributed replay buffer constitutes of many local buffers held per
        process, transmissions between processes only happen during sampling.

        During sampling, the tensors in "state", "action" and "next_state"
        dictionaries, along with "reward", will be concatenated in dimension 0.
        any other custom keys specified in ``**kwargs`` will not be
        concatenated.

        .. seealso:: :class:`.Buffer`

        Args:
            buffer_size: Maximum local buffer size.
            buffer_group: Process group which holds this buffer.
        """
        super(DistributedBuffer, self).__init__(buffer_size, "cpu")
        self.buffer_group = buffer_group
        self.buffer_group.rpc_register_paired(self.__class__, self)
        self.wr_lock = Lock()

    def size(self):
        """
        Returns:
            Length of current local buffer.
        """
        return len(self.buffer)

    def clear(self):
        # DOC INHERITED
        future = [
            self.buffer_group.rpc_paired_class_async(
                m, self._reply_clear, self.__class__
            )
            for m in self.buffer_group.get_group_members()
        ]
        results = [fut.wait() for fut in future]
        if not all(results):
            failed = [m for m, status
                      in zip(self.buffer_group.get_group_members(), results)
                      if not status]
            raise RuntimeError("Failed to perform clear on members {}"
                               .format(failed))

    def all_size(self):
        """
        Returns:
            Total length of all buffers.
        """
        future = []
        count = 0
        for m in self.buffer_group.get_group_members():
            future.append(self.buffer_group.rpc_paired_class_async(
                m, self.size, self.__class__
            ))
        for fut in future:
            count += fut.wait()
        return count

    def append(self, transition: Union[Transition, Dict],
               required_attrs=("state", "action", "next_state",
                               "reward", "terminal")):
        # DOC INHERITED
        self.wr_lock.acquire()
        super(DistributedBuffer, self).append(transition, required_attrs)
        self.wr_lock.release()

    def sample_batch(self,
                     batch_size: int,
                     concatenate: bool = True,
                     device: Union[str, t.device] = None,
                     sample_method: Union[Callable, str] = "random_unique",
                     sample_attrs: List[str] = None,
                     additional_concat_attrs: List[str] = None,
                     *_, **__) -> Any:
        # DOC INHERITED
        batch_size, batch = self._request_batch(batch_size, sample_method)

        if device is None:
            device = self.buffer_device
        if sample_attrs is None:
            sample_attrs = batch[0].keys()
        if additional_concat_attrs is None:
            additional_concat_attrs = []

        return \
            batch_size, \
            self.post_process_batch(batch, device, concatenate,
                                    sample_attrs, additional_concat_attrs)

    def _request_batch(self, batch_size: int,
                       sample_method: Union[Callable, str]) -> Any:
        """
        Used by samplers. Samplers will split the query equally to all buffer
        holders, and receive their reply.

        Args:
            batch_size: Size of batch to sample.
            sample_method: Sample method.

        Returns:
            Sampled batch size, batch data.
        """
        # TODO: add timeout
        p_num = self.buffer_group.size()
        local_batch_size = _round_up(batch_size / p_num)

        future = [
            self.buffer_group.rpc_paired_class_async(
                w, self._reply_batch, self.__class__,
                args=(local_batch_size, batch_size, sample_method)
            )
            for w in self._select_workers(batch_size)
        ]

        results = [fut.wait() for fut in future]
        all_batch_size = sum([r[0] for r in results])
        all_batch = list(it.chain([r[1] for r in results]))
        return all_batch_size, all_batch

    def _reply_batch(self, batch_size, sample_method):
        """
        Rpc function executed by buffer holders, will sample requested batch
        size.

        Args:
            batch_size: Size of batch to sample.
            sample_method: Sample method.

        Returns:
            Sampled batch size, batch data.
        """
        if isinstance(sample_method, str):
            if not hasattr(self, "sample_method_" + sample_method):
                raise RuntimeError("Cannot find specified sample method: {}"
                                   .format(sample_method))
            sample_method = getattr(self, "sample_method_" + sample_method)

        # sample raw local batch from local buffer

        self.wr_lock.acquire()
        local_batch_size, local_batch = sample_method(self.buffer,
                                                      batch_size)
        self.wr_lock.release()

        return local_batch_size, local_batch

    def _reply_clear(self, transition, priority, required_keys):
        self.wr_lock.acquire()
        super(DistributedBuffer, self).clear()
        self.wr_lock.release()
        return True

    def _select_workers(self, num: int):
        """
        Randomly select a sub group of workers.
        """
        workers = self.buffer_group.get_group_members().copy()
        np.random.shuffle(workers)
        return workers[:num]

    def __reduce__(self):
        raise RuntimeError("Distributed buffer is not pickable, "
                           "it is meant to be held per process!")



from typing import Union, Dict, List, Tuple, Callable, Any
import torch as t
import torch.nn as nn
import numpy as np

from machin.model.nets.base import NeuralNetworkModule
from machin.frame.buffers.buffer import Transition, Buffer
from .base import TorchFramework
from .utils import safe_call


class A2C(TorchFramework):
    """
    A2C framework.
    """

    _is_top = ["actor", "critic"]
    _is_restorable = ["actor", "critic"]

    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 learning_rate: float = 0.001,
                 entropy_weight: float = None,
                 value_weight: float = 0.5,
                 gradient_max: float = np.inf,
                 gae_lambda: float = 1.0,
                 discount: float = 0.99,
                 update_times: int = 50,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 visualize: bool = False,
                 **__):
        """
        Important:
            when given a state, and an optional, action actor must
            at least return two values:

            **1. Action**

              For **contiguous environments**, action must be of shape
              ``[batch_size, action_dim]`` and *clamped by action space*.
              For **discreet environments**, action could be of shape
              ``[batch_size, action_dim]`` if it is a one hot vector, or
              ``[batch_size, 1]`` if it is a categorically encoded integer.

            **2. Log likelihood of action (action probability)**

              For either type of environment, log likelihood is of shape
              ``[batch_size, 1]``.

              Action probability must be differentiable, Gradient of actor
              is calculated from the gradient of action probability.

            The third entropy value is optional:

            **3. Entropy of action distribution**

              Entropy is usually calculated using dist.entropy(), its shape
              is ``[batch_size, 1]``. You must specify ``entropy_weight``
              to make it effective.

        Hint:
            For contiguous environments, action's are not directly output by
            your actor, otherwise it would be rather inconvenient to calculate
            the log probability of action. Instead, your actor network should
            output parameters for a certain distribution
            (eg: :class:`~torch.distributions.categorical.Normal`)
            and then draw action from it.

            For discreet environments,
            :class:`~torch.distributions.categorical.Categorical` is sufficient,
            since differentiable ``rsample()`` is not needed.

            This trick is also known as **reparameterization**.

        Hint:
            Actions are from samples during training in the actor critic
            family (A2C, A3C, PPO, TRPO, IMPALA).

            When your actor model is given a batch of actions and states, it
            must evaluate the states, and return the log likelihood of the
            given actions instead of re-sampling actions.

            An example of your actor in contiguous environments::

                class ActorNet(nn.Module):
                    def __init__(self):
                        super(ActorNet, self).__init__()
                        self.fc = nn.Linear(3, 100)
                        self.mu_head = nn.Linear(100, 1)
                        self.sigma_head = nn.Linear(100, 1)

                    def forward(self, state, action=None):
                        x = t.relu(self.fc(state))
                        mu = 2.0 * t.tanh(self.mu_head(x))
                        sigma = F.softplus(self.sigma_head(x))
                        dist = Normal(mu, sigma)
                        action = (action
                                  if action is not None
                                  else dist.sample())
                        action_log_prob = dist.log_prob(action)
                        action_entropy = dist.entropy()
                        action = action.clamp(-2.0, 2.0)
                        return action.detach(), action_log_prob, action_entropy

        Hint:
            Entropy weight is usually negative, to increase exploration.

            Value weight is usually 0.5. So critic network converges less
            slowly than the actor network and learns more conditions.

            Update equation is equivalent to:

            :math:`Loss= w_e * Entropy + w_v * Loss_v + w_a * Loss_a`
            :math:`Loss_a = -log\\_likelihood * advantage`
            :math:`Loss_v = criterion(target\\_bellman\\_value - V(s))`

        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            gradient_max: Maximum gradient.
            gae_lambda: :math:`\\lambda` used in generalized advantage
                estimation.
            discount: :math:`\\gamma` used in the bellman function.
            update_times: Number of update iterations per sample period. Buffer
                will be cleared after ``update()``
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
        """
        self.update_times = update_times
        self.discount = discount
        self.value_weight = value_weight
        self.entropy_weight = entropy_weight
        self.grad_max = gradient_max
        self.gae_lambda = gae_lambda
        self.visualize = visualize

        self.actor = actor
        self.critic = critic
        self.actor_optim = optimizer(self.actor.parameters(),
                                     lr=learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(),
                                      lr=learning_rate)
        self.replay_buffer = (Buffer(replay_size, replay_device)
                              if replay_buffer is None
                              else replay_buffer)

        if lr_scheduler is not None:
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim,
                *lr_scheduler_args[0],
                **lr_scheduler_kwargs[0],
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim,
                *lr_scheduler_args[1],
                **lr_scheduler_kwargs[1]
            )

        self.criterion = criterion

        super(A2C, self).__init__()

    def act(self, state: Dict[str, Any], *_, **__):
        """
        Use actor network to give a policy to the current state.

        Returns:
            Anything produced by actor.
        """
        return safe_call(self.actor, state)

    def eval_act(self,
                 state: Dict[str, Any],
                 action: Dict[str, Any],
                 *_, **__):
        """
        Use actor network to evaluate the log-likelihood of a given
        action in the current state.

        Returns:
            Anything produced by actor.
        """
        return safe_call(self.actor, state, action)

    def criticize(self, state: Dict[str, Any], *_, **__):
        """
        Use critic network to evaluate current value.

        Returns:
            Value evaluated by critic.
        """
        return safe_call(self.critic, state)

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.

        Not suggested, since you will have to calculate "value" and "gae"
        by yourself.
        """
        self.replay_buffer.append(transition, required_attrs=(
            "state", "action", "next_state", "reward", "value",
            "gae", "terminal"
        ))

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.

        "value" and "gae" are automatically calculated.
        """
        if not isinstance(episode[0], Transition):
            episode = [Transition(**trans) for trans in episode]

        episode[-1]["value"] = episode[-1].reward

        # calculate value for each transition
        for i in reversed(range(1, len(episode))):
            episode[i - 1]["value"] = \
                episode[i]["value"] * self.discount + episode[i - 1]["reward"]

        # calculate advantage
        if self.gae_lambda == 1.0:
            for trans in episode:
                trans["gae"] = (trans["value"] -
                                self.criticize(trans["state"]).item())
        elif self.gae_lambda == 0.0:
            for trans in episode:
                trans["gae"] = (trans["reward"] +
                                self.discount * (1 - trans["terminal"])
                                * self.criticize(trans["next_state"]).item() -
                                self.criticize(trans["state"]).item())
        else:
            last_critic_value = 0
            last_gae = 0
            for trans in reversed(episode):
                critic_value = self.criticize(trans["state"]).item()
                gae_delta = (trans["reward"] +
                             self.discount * last_critic_value -
                             critic_value)
                last_critic_value = critic_value
                last_gae = trans["gae"] = (last_gae * self.discount
                                           * self.gae_lambda +
                                           gae_delta)

        for trans in episode:
            self.replay_buffer.append(trans, required_attrs=(
                "state", "action", "next_state", "reward", "value",
                "gae", "terminal"
            ))

    def update(self,
               update_value=True,
               update_policy=True,
               concatenate_samples=True,
               **__):
        """
        Update network weights by sampling from buffer. Buffer
        will be cleared after update is finished.

        Args:
            update_value: Whether update the Q network.
            update_policy: Whether update the actor network.
            concatenate_samples: Whether concatenate the samples.

        Returns:
            mean value of estimated policy value, value loss
        """

        sum_act_policy_loss = 0
        sum_value_loss = 0

        # sample a batch
        batch_size, (state, action, reward, next_state,
                     terminal, target_value, advantage, *others) = \
            self.replay_buffer.sample_batch(-1,
                                            sample_method="all",
                                            concatenate=concatenate_samples,
                                            sample_attrs=[
                                                "state", "action", "reward",
                                                "next_state", "terminal",
                                                "value", "gae", "*"],
                                            additional_concat_attrs=[
                                                "value", "gae"
                                            ])

        # normalize target value
        target_value = ((target_value - target_value.mean()) /
                        (target_value.std() + 1e-5))

        for _ in range(self.update_times):
            value = self.criticize(state)

            if self.entropy_weight is not None:
                new_action, new_action_log_prob, new_action_entropy = \
                    self.eval_act(state, action)
            else:
                new_action, new_action_log_prob, *_ = \
                    self.eval_act(state, action)
                new_action_entropy = None

            new_action_log_prob = new_action_log_prob.view(batch_size, 1)

            # calculate policy loss
            act_policy_loss = -(new_action_log_prob *
                                advantage.to(new_action_log_prob.device))

            if new_action_entropy is not None:
                act_policy_loss += (self.entropy_weight *
                                    new_action_entropy.mean())

            act_policy_loss = act_policy_loss.mean()

            value_loss = (self.criterion(target_value.to(value.device), value) *
                          self.value_weight)

            if self.visualize:
                self.visualize_model(act_policy_loss, "actor")

            # Update actor network
            if update_policy:
                self.actor.zero_grad()
                act_policy_loss.backward()
                nn.utils.clip_grad_norm_(
                    self.actor.parameters(), self.grad_max
                )
                self.actor_optim.step()
                sum_act_policy_loss += act_policy_loss.item()

            if self.visualize:
                self.visualize_model(value_loss, "critic")

            # Update critic network
            if update_value:
                self.critic.zero_grad()
                value_loss.backward()
                nn.utils.clip_grad_norm_(
                    self.critic.parameters(), self.grad_max
                )
                self.critic_optim.step()
                sum_value_loss += value_loss.item()

        self.replay_buffer.clear()
        return (-sum_act_policy_loss / self.update_times,
                sum_value_loss / self.update_times)

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()



from multiprocessing import get_context, get_start_method
from . import distributed, server, assigner, pickle, pool, queue

__all__ = [
    "get_context",
    "get_start_method",
    "distributed",
    "server",
    "assigner",
    "pickle",
    "pool",
    "queue"
]



from machin.frame.buffers.prioritized_buffer import PrioritizedBuffer
from machin.utils.logging import default_logger
# pylint: disable=wildcard-import, unused-wildcard-import
from .ddpg import *


class DDPGPer(DDPG):
    """
    DDPG with prioritized experience replay.

    Warning:
        Your criterion must return a tensor of shape ``[batch_size,1]``
        when given two tensors of shape ``[batch_size,1]``, since we
        need to multiply the loss with importance sampling weight
        element-wise.

        If you are using loss modules given by pytorch. It is always
        safe to use them without any modification.
    """
    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 actor_target: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 reward_func: Callable = None,
                 action_trans_func: Callable = None,
                 visualize: bool = False,
                 **__):
        # DOC INHERITED
        super(DDPGPer, self).__init__(
            actor, actor_target, critic, critic_target, optimizer, criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            update_rate=update_rate,
            learning_rate=learning_rate,
            discount=discount,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=(PrioritizedBuffer(replay_size, replay_device)
                           if replay_buffer is None
                           else replay_buffer),
            reward_func=reward_func,
            action_trans_func=action_trans_func,
            visualize=visualize
        )
        # reduction must be None
        if not hasattr(self.criterion, "reduction"):
            raise RuntimeError("Criterion must have the 'reduction' property")
        else:
            if hasattr(self.criterion, "reduction"):
                # A loss defined in ``torch.nn.modules.loss``
                if self.criterion.reduction != "none":
                    default_logger.warning(
                        "The reduction property of criterion is not 'none', "
                        "automatically corrected."
                    )
                    self.criterion.reduction = "none"

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        (batch_size,
         (state, action, reward, next_state, terminal, *others),
         index, is_weight) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        # Update critic network first.
        # Generate value reference :math: `y_i` using target actor and
        # target critic.
        with t.no_grad():
            next_action = self.action_transform_func(self.act(next_state, True),
                                                     next_state, *others)
            next_value = self.criticize(next_state, next_action, True)
            next_value = next_value.view(batch_size, -1)
            y_i = self.reward_func(reward, self.discount, next_value,
                                   terminal, *others)

        # critic loss
        cur_value = self.criticize(state, action)
        value_loss = self.criterion(cur_value, y_i.to(cur_value.device))
        value_loss = (value_loss *
                      t.from_numpy(is_weight).view([batch_size, 1])
                      .to(value_loss.device))
        value_loss = value_loss.mean()

        if self.visualize:
            self.visualize_model(value_loss, "critic")

        # actor loss
        cur_action = self.action_transform_func(self.act(state), state, *others)
        act_value = self.criticize(state, cur_action)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        # update priority
        abs_error = t.sum(t.abs(act_value - y_i), dim=1).flatten().cpu().numpy()
        self.replay_buffer.update_priority(abs_error, index)

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor")

        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            self.critic_optim.step()

        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            self.actor_optim.step()

        # Update target networks
        if update_target:
            soft_update(self.actor_target, self.actor, self.update_rate)
            soft_update(self.critic_target, self.critic, self.update_rate)

        # use .item() to prevent memory leakage
        return -act_policy_loss.item(), value_loss.item()



from random import choice
from typing import Union, Dict, List, Tuple, Any
from threading import Lock
from ..transition import Transition
from .prioritized_buffer import PrioritizedBuffer
from machin.parallel.distributed import RpcGroup
import numpy as np
import torch as t


class DistributedPrioritizedBuffer(PrioritizedBuffer):
    def __init__(self, buffer_size: int, buffer_group: RpcGroup, *_, **__):
        """
        Create a distributed prioritized replay buffer instance.

        To avoid issues caused by tensor device difference, all transition
        objects are stored in device "cpu".

        Distributed prioritized replay buffer constitutes of many local buffers
        held per process, since it is very inefficient to maintain a weight
        tree across processes, each process holds a full copy of the weight
        tree and only a part of the buffer (same as ``DistributedBuffer``),
        transmissions between processes happen during appending, updating and
        sampling.

        During sampling, the tensors in "state", "action" and "next_state"
        dictionaries, along with "reward", will be concatenated in dimension 0.
        any other custom keys specified in ``**kwargs`` will not be
        concatenated.

        .. seealso:: :class:`PrioritizedBuffer`

        Args:
            buffer_size: Maximum local buffer size.
            buffer_group: Process group which holds this buffer.
        """
        super(DistributedPrioritizedBuffer, self) \
            .__init__(buffer_size, "cpu")
        self.buffer_group = buffer_group
        self.buffer_group.rpc_register_paired(self.__class__, self)
        self.wr_lock = Lock()

    def append(self,
               transition: Union[Transition, Dict],
               priority: Union[float, None] = None,
               required_attrs=("state", "action", "next_state",
                               "reward", "terminal")):
        # DOC INHERITED
        # TODO: batched append
        future = [
            self.buffer_group.rpc_paired_class_async(
                m, self._reply_append, self.__class__,
                args=(transition, priority, required_attrs)
            )
            for m in self.buffer_group.get_group_members()
        ]
        results = [fut.wait() for fut in future]
        if not all(results):
            failed = [m for m, status
                      in zip(self.buffer_group.get_group_members(), results)
                      if not status]
            raise RuntimeError("Failed to perform append on members {}"
                               .format(failed))

    def size(self):
        """
        Returns:
            Length of current local buffer.
        """
        return len(self.buffer)

    def all_size(self):
        """
        Returns:
            Total length of all buffers.
        """
        future = []
        count = 0
        for m in self.buffer_group.get_group_members():
            future.append(self.buffer_group.rpc_paired_class_async(
                m, self.size, self.__class__
            ))
        for fut in future:
            count += fut.wait()
        return count

    def clear(self):
        # DOC INHERITED
        future = [
            self.buffer_group.rpc_paired_class_async(
                m, self._reply_clear, self.__class__
            )
            for m in self.buffer_group.get_group_members()
        ]
        results = [fut.wait() for fut in future]
        if not all(results):
            failed = [m for m, status
                      in zip(self.buffer_group.get_group_members(), results)
                      if not status]
            raise RuntimeError("Failed to perform append on members {}"
                               .format(failed))

    def update_priority(self, priorities: np.ndarray, indexes: np.ndarray):
        # DOC INHERITED
        future = [
            self.buffer_group.rpc_paired_class_async(
                m, self._reply_update_priority, self.__class__,
                args=(priorities, indexes)
            )
            for m in self.buffer_group.get_group_members()
        ]
        results = [fut.wait() for fut in future]
        if not all(results):
            failed = [m for m, status
                      in zip(self.buffer_group.get_group_members(), results)
                      if not status]
            raise RuntimeError("Failed to perform update priority on members {}"
                               .format(failed))

    def sample_batch(self,
                     batch_size: int,
                     concatenate: bool = True,
                     device: Union[str, t.device] = None,
                     sample_attrs: List[str] = None,
                     additional_concat_attrs: List[str] = None,
                     *_, **__) -> Tuple[Any, Any, np.ndarray, np.ndarray]:
        # DOC INHERITED
        worker = choice(self.buffer_group.get_group_members())
        return self.buffer_group.rpc_paired_class_async(
            worker, self._reply_sample, self.__class__,
            args=(batch_size, concatenate, device,
                  sample_attrs, additional_concat_attrs)
        )

    def _reply_clear(self):
        self.wr_lock.acquire()
        super(DistributedPrioritizedBuffer, self).clear()
        self.wr_lock.release()
        return True

    def _reply_update_priority(self, priorities, indexes):
        self.wr_lock.acquire()
        super(DistributedPrioritizedBuffer, self).update_priority(priorities,
                                                                  indexes)
        self.wr_lock.release()
        return True

    def _reply_append(self, transition, priority, required_keys):
        self.wr_lock.acquire()
        super(DistributedPrioritizedBuffer, self).append(transition, priority,
                                                         required_keys)
        self.wr_lock.release()
        return True

    def _reply_sample(self, batch_size, concatenate, device,
                      sample_keys, additional_concat_keys):
        self.wr_lock.acquire()
        result = super(DistributedPrioritizedBuffer, self).sample_batch(
            batch_size, concatenate, device, sample_keys, additional_concat_keys
        )
        self.wr_lock.release()
        return result



import gym
import numpy as np
from itertools import repeat
from typing import Tuple
from ..utils.openai_gym import disable_view_window
from machin.parallel.pool import Pool
from machin.parallel import get_context
from .base import *

disable_view_window()


def _call_gym_env_method(rank, all_env_idx, method, args=(), kwargs=None):
    if kwargs is None:
        kwargs = {}
    hold = set(_init_envs.env_idxs[rank])
    use = hold.intersection(set(all_env_idx))
    for idx in use:
        env = _init_envs.envs[rank][idx]
        if hasattr(env, method):
            return getattr(env, method)(*args, **kwargs)


def _init_envs(rank, envs, env_idxs):
    _init_envs.envs = dict()
    _init_envs.env_idxs = dict()
    # use rank to avoid overwriting when using ThreadPool
    _init_envs.envs[rank] = {i: e for i, e in zip(env_idxs, envs)}
    _init_envs.env_idxs[rank] = env_idxs


class GymTerminationError(Exception):
    def __init__(self):
        super(GymTerminationError, self).__init__(
            "One or several environments have terminated, "
            "reset before continuing."
        )


class ParallelWrapperDummy(ParallelWrapperBase):
    """
    Dummy parallel wrapper for gym environments, implemented using for-loop.

    For debug purpose only.
    """
    def __init__(self, envs: List[gym.Env]):
        """
        Args:
            envs: List of gym environments.
        """
        super(ParallelWrapperDummy, self).__init__()
        self._envs = envs
        self._terminal = np.zeros([len(self._envs)], dtype=np.bool)

    def reset(self, idx: Union[int, List[int]] = None) -> np.ndarray:
        """
        Returns:
            Batched gym states, the first dimension is environment size.
        """
        if idx is None:
            obsrv = np.stack([e.reset() for e in self._envs])
            self._terminal = np.zeros([self.size()], dtype=np.bool)
        else:
            obsrv = []
            if np.isscalar(idx):
                idx = [idx]
            for i in idx:
                obsrv.append(self._envs[i].reset())
                self._terminal[i] = False
        return obsrv

    def step(self, action: np.ndarray, idx: Union[int, List[int]] = None) \
            -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Let specified environment(s) run one time step. Specified environments
        must be active and have not reached terminal states before.

        Args:
            action: Actions sent to each specified environment, the size of the
                first dimension must match the number of selected environments.
            idx: Indexes of selected environments, default is all.

        Returns:
            Stacked observation, reward, terminal, info in ``np.ndarray``
        """
        if idx is None:
            idx = list(range(self.size()))
        elif np.isscalar(idx):
            idx = [idx]

        if len(action) != len(idx):
            raise ValueError("Action number must match environment number!")
        if np.any(self._terminal[idx]):
            raise GymTerminationError()

        envs = [self._envs[i] for i in idx]
        result = [e.step(a) for e, a in zip(envs, action)]
        obsrv, reward, terminal, info = zip(*result)

        obsrv = np.stack(obsrv)
        reward = np.stack(reward)
        terminal = np.stack(terminal)
        info = np.stack(info)

        self._terminal[idx] |= terminal

        return obsrv, reward, terminal, info

    def seed(self, seed: Union[int, List[int]] = None) -> List[int]:
        """
        Set seeds for all environments.

        Args:
            seed: If seed is ``int``, the same seed will be used for all
                environments.
                If seed is ``List[int]``, it must have the same size as
                the number of all environments.
                If seed is ``None``, all environments will use the default
                seed.

        Returns:
            Actual used seed returned by all environments.
        """
        if np.isscalar(seed) or seed is None:
            seed = [seed] * self.size()
        result = []
        for e, s in zip(self._envs, seed):
            if hasattr(e, 'seed'):
                result.append(e.seed(s))
        return result

    def render(self, idx: Union[int, List[int]] = None,
               *_, **__) -> List[np.ndarray]:
        """
        Render all/specified environments.

        Args:
            idx: Indexes of selected environments, default is all.

        Returns:
            A list or rendered frames, of type ``np.ndarray`` and
            size (H, W, 3).
        """
        rendered = []
        if idx is None:
            for e in self._envs:
                if np.any(self._terminal):
                    raise GymTerminationError()
                rendered.append(e.render(mode="rgb_array"))
        else:
            if np.isscalar(idx):
                idx = [idx]
            for i in idx:
                if self._terminal[i]:
                    raise GymTerminationError()
                rendered.append(self._envs[i].render(mode="rgb_array"))
        return rendered

    def close(self) -> None:
        """
        Close all environments.
        """
        for e in self._envs:
            e.close()

    def active(self) -> List[int]:
        """
        Returns: Indexes of current active environments.
        """
        return np.arange(self.size())[~self._terminal]

    def size(self) -> int:
        """
        Returns: Number of environments.
        """
        return len(self._envs)


class ParallelWrapperPool(ParallelWrapperBase):
    """
    Parallel wrapper based on thread pool or subprocess pool for gym
    environments. Environments are passed to processes/threads in pool
    during ``__init__()``, further operations on environments will only
    use indexes of environments.
    """
    def __init__(self, envs: List[gym.Env], pool=None, pool_size=None) -> None:
        """
        Args:
            envs: List of gym environments.
            pool: A pool of workers, by default it is a subprocess pool.
            pool_size: Number of workers in the pool
        """
        super(ParallelWrapperPool, self).__init__()

        if pool is not None and pool_size is not None:
            self.pool = pool
            self.pool_size = pool_size
        else:
            ctx = get_context("spawn")
            pool = Pool(context=ctx, is_daemon=False, is_global=False)
            self.pool = pool
            self.pool_size = pool_size = pool.size()

        self.env_size = env_size = len(envs)

        # delegate environments to workers
        chunk_size = int(np.ceil(env_size / pool_size))
        env_idxs = list(range(env_size))
        env_chunks = [envs[i:i + chunk_size]
                      for i in range(0, env_size, chunk_size)]
        env_idx_chunks = [env_idxs[i:i + chunk_size]
                          for i in range(0, env_size, chunk_size)]

        # initialize the pool, store environments and their indexes
        # as a member of function _init_envs()
        self.pool.starmap(_init_envs,
                          zip(range(pool_size), env_chunks, env_idx_chunks))
        self._terminal = np.zeros([env_size], dtype=np.bool)

    def reset(self, idx: Union[int, List[int]] = None) -> np.ndarray:
        """
        Returns:
            Batched gym states, the first dimension is environment size.
        """
        env_idxs = self._select_envs(idx)
        self._terminal[env_idxs] = False
        obsrv = np.stack(
            self.pool.starmap(_call_gym_env_method,
                              zip(*self._process(env_idxs), repeat("reset")))
        )
        return obsrv

    def step(self, action: np.ndarray, idx: Union[int, List[int]] = None) \
            -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Let specified environment(s) run one time step. Specified environments
        must be active and have not reached terminal states before.

        Args:
            action: Actions sent to each specified environment, the size of the
                first dimension must match the number of selected environments.
            idx: Indexes of selected environments, default is all.

        Returns:
            observation, reward, terminal, info
        """
        env_idxs = self._select_envs(idx)
        if len(action) != len(env_idxs):
            raise ValueError("Action number must match environment number!")

        result = self.pool.starmap(
            _call_gym_env_method,
            zip(*self._process(env_idxs), repeat("step"), action)
        )
        obsrv, reward, terminal, info = zip(*result)

        obsrv = np.stack(obsrv)
        reward = np.stack(reward)
        terminal = np.stack(terminal)
        info = np.stack(info)

        self._terminal[env_idxs] |= terminal

        return obsrv, reward, terminal, info

    def seed(self, seed: Union[int, List[int]] = None) -> List[int]:
        """
        Set seeds for all environments.

        Args:
            seed: If seed is ``int``, the same seed will be used for all
                environments.
                If seed is ``List[int]``, it must have the same size as
                the number of all environments.
                If seed is ``None``, all environments will use the default
                seed.

        Returns:
            Actual used seed returned by all environments.
        """
        if np.isscalar(seed) or seed is None:
            seed = [seed] * self.size()
        env_idxs = self._select_envs()
        result = self.pool.starmap(
            _call_gym_env_method,
            zip(*self._process(env_idxs), repeat("seed"), seed)
        )
        return result

    def render(self, idx: Union[int, List[int]] = None,
               *args, **kwargs) -> List[np.ndarray]:
        """
        Render all/specified environments.

        Args:
            idx: Indexes of selected environments, default is all.

        Returns:
            A list or rendered frames, of type ``np.ndarray`` and size
            (H, W, 3).
        """
        env_idxs = self._select_envs(idx)
        rendered = self.pool.starmap(
            _call_gym_env_method,
            zip(*self._process(env_idxs),
                repeat("render"),
                repeat(()),
                repeat({"mode": "rgb_array"}))
        )
        return rendered

    def close(self) -> None:
        """
        Close all environments.
        """
        env_idxs = self._select_envs()
        self.pool.starmap(
            _call_gym_env_method,
            zip(*self._process(env_idxs), repeat("close"))
        )

    def active(self) -> List[int]:
        """
        Returns: Indexes of current active environments.
        """
        return np.arange(self.size())[~self._terminal]

    def size(self) -> int:
        """
        Returns: Number of environments.
        """
        return self.env_size

    def _select_envs(self, idx=None):
        if idx is None:
            idx = list(range(self.env_size))
        else:
            if np.isscalar(idx):
                idx = [idx]
        return idx

    def _process(self, idx):
        return range(self.pool_size), idx


class ParallelWrapperGroup(ParallelWrapperBase):
    # TODO: implement parallel wrapper based on process group
    def reset(self, idx: Union[int, List[int], None] = None) -> Any:
        raise NotImplementedError

    def step(self, action, idx: Union[int, List[int], None] = None) -> Any:
        raise NotImplementedError

    def seed(self, seed: Union[int, List[int], None] = None) -> List[int]:
        raise NotImplementedError

    def render(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def close(self) -> Any:
        raise NotImplementedError

    def active(self) -> List[int]:
        raise NotImplementedError

    def size(self) -> int:
        raise NotImplementedError



from typing import Callable, Tuple, Any
import torch as t
import torch.nn as nn

from machin.utils.logging import default_logger as logger
from machin.utils.helper_classes import Switch
from .generator import NormalNoiseGen


class AdaptiveParamNoise(object):
    def __init__(self,
                 initial_stddev: float = 0.1,
                 desired_action_stddev: float = 0.1,
                 adoption_coefficient: float = 1.01):
        """
        Implements the adaptive parameter space method in
        `<<Parameter space noise for exploration>> \
<https://arxiv.org/pdf/1706.01905.pdf>`_.

        Hint:
            Let :math:`\\theta` be the standard deviation of noise,
            and :math:`\\alpha` be the adpotion coefficient, then:

            :math:`\\theta_{n+1} = \\left \\{ \
                \\begin{array}{ll} \
                    \\alpha \\theta_k \
                        & if\\ d(\\pi,\\tilde{\\pi})\\leq\\delta, \\\\ \
                    \\frac{1}{\\alpha} \\theta_k & otherwise, \
                \\end{array} \
            \\right. \\ `

            Noise is directly applied to network parameters.

        Args:
            initial_stddev: Initial noise standard deviation.
            desired_action_stddev: Desired standard deviation for
            adoption_coefficient: Adoption coefficient.
        """

        self.initial_stddev = initial_stddev
        self.desired_action_stddev = desired_action_stddev
        self.adoption_coefficient = adoption_coefficient

        self.current_stddev = initial_stddev

    def adapt(self, distance: float):
        """
        Update noise standard deviation according to distance.

        Args:
            distance: Current distance between the noisy action and clean
                action.
        """
        if distance > self.desired_action_stddev:
            # Decrease stddev.
            self.current_stddev /= self.adoption_coefficient
        else:
            # Increase stddev.
            self.current_stddev *= self.adoption_coefficient

    def get_dev(self) -> float:
        """
        Returns:
            Current noise standard deviation.
        """
        return self.current_stddev

    def __repr__(self):
        fmt = 'AdaptiveParamNoise(i_std={}, da_std={}, adpt_coeff={})'
        return fmt.format(self.initial_stddev,
                          self.desired_action_stddev,
                          self.adoption_coefficient)


def _gen_perturb_hook(module, perturb_switch, reset_switch, perturb_gen):
    org_params = {}
    noisy_params = {}

    def perturb_pre_hook(*_):
        with t.no_grad():
            if perturb_switch.get():
                if noisy_params and not reset_switch.get():
                    # Use generated noisy parameters.
                    for param_name, param_value in module.named_parameters():
                        param_value.set_(noisy_params[param_name])
                else:
                    # Generate noisy parameters if they don't exist.
                    org_params.clear()
                    noisy_params.clear()
                    for param_name, param_value in module.named_parameters():
                        org_params[param_name] = param_value.clone()
                        param_value += perturb_gen(param_value.shape,
                                                   param_value.device).detach()
                        noisy_params[param_name] = param_value.clone()

            elif not perturb_switch.get():
                # Use original parameters
                if org_params:
                    for param_name, param_value in module.named_parameters():
                        param_value.set_(org_params[param_name])

    def perturb_post_hook(*_):
        # Called before backward update, swap noisy parameters out,
        # so gradients are applied to original parameters.
        with t.no_grad():
            if org_params:
                for param_name, param_value in module.named_parameters():
                    param_value.set_(org_params[param_name])

    return perturb_pre_hook, perturb_post_hook


def perturb_model(model: nn.Module,
                  perturb_switch: Switch,
                  reset_switch: Switch,
                  distance_func: Callable =
                  lambda x, y: t.dist(x, y, 2).mean().item(),
                  desired_action_stddev: float = 0.5,
                  noise_generator: Any = NormalNoiseGen,
                  noise_generator_args: Tuple = (),
                  noise_generate_function: Callable = None):
    """
    Give model's parameters a little perturbation.

    Hint:
        1. ``noise_generator`` must accept (shape, \*args) in its ``__init__``
        function, where shape is the required shape. it also needs to have
        ``__call__(device=None)`` which produce a noise tensor on the specified
        device when invocated.

        2. ``noise_generate_function`` must accept (shape, device, std:float)
        and return a noise tensor on the specified device.

    Example:
        In order to use this function to perturb your model, you need to::

            from machin.utils.helper_classes import Switch
            from machin.frame.noise.param_space_noise import perturb_model
            from machin.utils.visualize import visualize_graph
            import torch as t

            dims = 5

            t.manual_seed(0)
            model = t.nn.Linear(dims, dims)
            optim = t.optim.Adam(model.parameters(), 1e-3)
            p_switch, r_switch = Switch(), Switch()
            rst_func = perturb_model(model, p_switch, r_switch)
            r_switch.on()

            # turn off/on the perturbation switch to see the difference
            p_switch.on()

            # do some sampling
            action = model(t.ones([dims]))

            # Visualize will not show any leaf noise tensors
            # because they are created in t.no_grad() context.
            visualize_graph(action, exit_after_vis=False)

            # do some training
            loss = (action - t.ones([dims])).sum()
            loss.backward()
            rst_func()
            optim.step()
            print(model.weight)

    Args:
        model: Neural network model.
        perturb_switch: The switch used to enable perturbation. If switch is
            set to ``False`` (off), then during the forward process, original
            parameters are used.
        reset_switch: The switch used to reset perturbation noise. If switch is
            set to ``True`` (on), and ``perturb_switch`` is also on, then during
            every forward process, a new set of noise is applied to each param.
            If only ``perturb_switch`` is on, then the same set of noise is used
            in the forward process.
        distance_func: Distance function, accepts two tensors produced by
            ``model`` (one is noisy), return the distance as float.
        desired_action_stddev: Desired action standard deviation.
        noise_generator: Noise generator class, Ornstein-Uhlenbeck noise
            generator is not supported because it has internal state.
        noise_generator_args: Additional args other than shape of the noise
            generator.
        noise_generate_function: Noise generation function, mutually exclusive
            with ``noise_generator`` and ``noise_generator_args``.

    Returns:
        A reset function with no arguments, will swap out noisy parameters and
        swap in clean parameters when called.
    """
    tmp_action = {}
    post_hooks = []
    param_noise_spec = AdaptiveParamNoise(
        desired_action_stddev=desired_action_stddev
    )

    def param_noise_gen(shape, device):
        gen = noise_generator(shape, *noise_generator_args)
        return gen(device) * param_noise_spec.get_dev()

    def param_noise_custom_gen_wrapper(shape, device):
        std_dev = param_noise_spec.get_dev()
        return noise_generate_function(shape, device, std_dev)

    if noise_generate_function is not None:
        param_noise_gen = param_noise_custom_gen_wrapper

    def perturb_adjust_hook(_model, _input, output):
        if perturb_switch.get():
            tmp_action["with_noise"] = output.clone()
        else:
            tmp_action["without_noise"] = output.clone()
        if "with_noise" in tmp_action and "without_noise" in tmp_action:
            # compute distance
            with t.no_grad():
                dist = distance_func(tmp_action["with_noise"],
                                     tmp_action["without_noise"])
                tmp_action.clear()
                param_noise_spec.adapt(dist)
                logger.info("Current output distance: {}".format(dist))
                logger.info("Current param noise stddev: {}"
                            .format(param_noise_spec.get_dev()))

    model.register_forward_hook(perturb_adjust_hook)

    for _sub_name, sub_module in model.named_modules():
        sub_module = sub_module  # type: nn.Module
        if len([i for i in sub_module.modules()]) != 1:
            continue
        pre_f, post = _gen_perturb_hook(sub_module, perturb_switch,
                                        reset_switch, param_noise_gen)
        sub_module.register_forward_pre_hook(pre_f)
        post_hooks.append(post)

    def reset():
        for h in post_hooks:
            h()

    return reset



from . import ordered_server
from . import param_server
from .ordered_server import SimpleOrderedServer
from .param_server import PushPullGradServer, PushPullModelServer

__all__ = ["SimpleOrderedServer",
           "PushPullGradServer",
           "PushPullModelServer",
           "ordered_server", "param_server"]



import os
import time
from torch.multiprocessing import Process
import torch.distributed.rpc as rpc

rank = None
messages = 1000
world_size = 4


def test():
    print("Process {} received request.".format(rank))
    time.sleep(1)
    return "SPAM"


def run(i):
    global rank
    rank = i
    if i == world_size - 1:
        time.sleep(4)
        print("Process {} delayed start.".format(i))
    rpc.init_rpc("Rank" + str(i), rank=i, world_size=world_size,
                 rpc_backend_options=rpc.ProcessGroupRpcBackendOptions(
                     init_method="env://",
                     rpc_timeout=rpc.timedelta(seconds=2),
                     num_send_recv_threads=4))
    if i == 0:
        time.sleep(2)
        print("Process 0 exit.")
        exit(-1)
    t = time.time()
    reqs = []
    for j in range(messages):
        for r in range(world_size):
            reqs.append(rpc.rpc_async("Rank{}".format(r), test, args=()))
        for req, idx in zip(reqs, range(world_size)):
            try:
                print("{} Received from {} : {}".format(rank, idx, req.wait()))
            except RuntimeError:
                print("An error ocurred while {} receiving results from {}".format(rank, idx))
        reqs.clear()
    print(time.time() - t)
    rpc.shutdown(graceful=False)


if __name__ == "__main__":
    os.environ['MASTER_ADDR'] = "localhost"
    os.environ['MASTER_PORT'] = "29500"
    ps = []
    for i in range(world_size):
        p = Process(target=run, args=(i,))
        p.start()
        ps.append(p)

    for p in ps:
        p.join()
